;
; assembler
;

; (in-package "assembler")

; (export 'assemble)

; (export '(add.b add.w add.l ...))



; ---------------------------------------------------------------------------

;;xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
;;
;;			compatibility stuff
;;
;;xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

(define (last-pair l)
  (define (last-aux l)
    (if (atom? (cdr l)) l (last-aux (cdr l))))
  (if (atom? l) l (last-aux l)))

(define **prop-list** '())

(define (prop-set! id p val)
  (let ((x (assq id **prop-list**)))
    (if (not x)
      (set! **prop-list** (cons (cons id (list (cons p val))) **prop-list**))
      (let ((y (assq p (cdr x))))
        (if (not y)
          (set-cdr! x (cons (cons p val) (cdr x)))
          (set-cdr! y val))))))

(define (prop-ref id p)
  (let ((x (assq id **prop-list**)))
    (if (atom? x)
      #!false
      (let ((y (assq p (cdr x))))
        (if (atom? y)
          #!false
          (cdr y))))))


;;xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

(define (proper-list l)
  (if (pair? l) (proper-list (cdr l)) (null? l)))

(define (tlist . l)   ; construct a tail list containing the values in l
  (let ((x (cons '() l)))
    (set-car! x (last-pair x))
    x))

(define (tlist-add l x)   ; add value x to end of tail list l
                          ; returns the last pair of the new tail list
  (let ((tail (car l)))
    (set-car! l (list x))
    (set-cdr! tail (car l))
    (car l)))

(define (tlist->list l)   ; returns the list of the values contained in tail list l
  (cdr l))

(define (tlist-last-pair l)   ; returns the last pair in tail list l
  (last-pair l))

(define (insert! l x)
  (let ((y (cons x (cdr l))))
    (set-cdr! l y)
    y))

(define **dynamic-environment** '())

(macro bind
  (lambda (l)
    (let ((var (cadr l)) (val (caddr l)) (form (cadddr l)))
      `(begin (set! **dynamic-environment**
                    (cons (cons ',var ,val) **dynamic-environment**))
              (let ((x ,form))
                (set! **dynamic-environment**
                      (cdr **dynamic-environment**))
                x)))))

(macro value
  (lambda (l)
    `(cdr (assq ',(cadr l) **dynamic-environment**))))

(macro set-value!
  (lambda (l)
    `(set-cdr! (assq ',(cadr l) **dynamic-environment**) ,(caddr l))))

(define (prin1-hex x)
  (define (print-hex-digit x)
    (write-char (string-ref "0123456789ABCDEF" (logand x 15))))
  (if (integer? x)
    (let ((y (if (< x 0) (+ x 65536) x)))
      (print-hex-digit (lsh y -12))
      (print-hex-digit (lsh y -8))
      (print-hex-digit (lsh y -4))
      (print-hex-digit y))
    (write x)))

(define (print-hex x)
  (prin1-hex x)
  (newline))

; ---------------------------------------------------------------------------

(define (s8bit? n)
  (and (>= n -128) (<= n 127)))

(define (s16bit? n)
  (and (>= n -32768) (<= n 32767)))

(define (u8bit? n)
  (and (>= n -128) (<= n 255)))

(define (u16bit? n)
  (and (>= n -32768) (<= n 65535)))

;;;_____________________________________________________________________________
;;;
;;; FIXUPS
;;;

(define (make-fixups)
  (tlist))

(define (fixups-add fixs fix)
  (tlist-add fixs fix))

(define (fixups->list fixs)
  (tlist->list fixs))

;; lbl fixup:

(define (make-lbl)
  (vector 0 0))

(define (lbl-pass lbl)
  (vector-ref lbl 0))

(define (lbl-pass-set! lbl pass)
  (vector-set! lbl 0 pass))

;; reference fixup:

(define (make-ref kind loc lbl pos)
  (vector kind loc lbl pos))

(define (ref-kind ref)
  (vector-ref ref 0))

(define (ref-kind-set! ref kind)
  (vector-set! ref 0 kind))

(define (ref-pos ref)
  (vector-ref ref 3))

(define (ref-dist ref)
  (- (location (vector-ref ref 2)) (location ref)))

(define (ref-dist-pass ref pass ofs)
  (if (= (lbl-pass (vector-ref ref 2)) pass)
    (+ (ref-dist ref) ofs)
    (ref-dist ref)))

;; general fixup manipulation:

(define (lbl? fix)
  (integer? (vector-ref fix 0)))

(define (location fix)
  (vector-ref fix 1))

(define (location-set! fix loc)
  (vector-set! fix 1 loc))

(define (fixup fixups)
  (define (resolve l pass ofs)
    (define (resolve-aux fix)
      (location-set! fix (+ (location fix) ofs))
      (cond ((lbl? fix)
             (lbl-pass-set! fix pass))
            ((eq? (ref-kind fix) 'bra)
             (let ((d (ref-dist-pass fix pass ofs)))
               (if (or (zero? d) (not (s8bit? d)))
                 (begin (ref-kind-set! fix '16bit) (set! ofs (+ ofs 2))))))))
    (for-each resolve-aux l)
    (if (> ofs 0) (resolve l (1+ pass) 0)))
  (define (patch fix)
    (if (not (lbl? fix))
      (let ((d (ref-dist fix)))
        (if (eq? (ref-kind fix) '16bit)
          (if (s16bit? d) (insert! (ref-pos fix) d)
                          (error "16 bit displacement out of range"))
          (if (s8bit? d) (set-car! (ref-pos fix)
                           (+ (car (ref-pos fix))
                              (if (negative? d) (+ d 256) d)))
                         (error "8 bit displacement out of range"))))))
  (resolve (fixups->list fixups) 1 0)
  (for-each patch (fixups->list fixups)))


;;;_____________________________________________________________________________
;;;
;;; LABELS
;;;

(define (make-labels)
  (tlist))

(define (labels-add lbls name lbl)
  (if (assq name (tlist->list lbls))
    (error "asm: duplicate label " name)
    (tlist-add lbls (cons name lbl))))

(define (labels-get lbls name)
  (let ((x (assq name (tlist->list lbls))))
    (if x
      (cdr x)
      (error "asm: undefined label " name))))

;;;_____________________________________________________________________________
;;;
;;; CODES
;;;

(define (make-codes)
  (tlist))

(define (codes-add cods cod)
  (tlist-add cods cod))

(define (codes-last-pair cods)
  (tlist-last-pair cods))

(define (codes->list cods)
  (tlist->list cods))


;;;_____________________________________________________________________________
;;;
;;; EMIT
;;;

(define (emit-16bit x)
  (set-value! *location* (+ (value *location*) 2))
  (codes-add (value *codes*) x))

(define (emit-ref kind lbl ofs)
  (fixups-add (value *fixups*)
    (make-ref kind (- (- (value *location*) ofs) (if (eq? kind '8bit) 2 0))
              lbl (codes-last-pair (value *codes*))))
  (if (eq? kind '16bit) (set-value! *location* (+ (value *location*) 2))))

(define (emit-label name)
  (let ((lbl (labels-get (value *labels*) name)))
    (location-set! lbl (value *location*))
    (fixups-add (value *fixups*) lbl)))

(define (emit-quote val)
  (set-value! *location* (+ (value *location*) 4))
  (codes-add (value *codes*) (list 'quote val)))


;;;_____________________________________________________________________________
;;;
;;; ASSEMBLE
;;;


(define (quoted? x)
  (and (couple x) (eq? (car x) 'quote)))

(define (opcode x)
  (let ((y (prop-ref (car x) 'asm)))
    (if y
      (let ((z (length (cdr x))))
        (if (or (= z (cddr y)) (negative? (cddr y)))
          (case z
            ((0) ((cadr y) (car y)))
            ((1) ((cadr y) (car y) (cadr x)))
            ((2) ((cadr y) (car y) (cadr x) (caddr x))))
          (err "incorrect number of operands")))
      (err "unknown opcode"))))

(define (label? x)
  (and (symbol? x) (not (r? x))))

(define (pass1 x)
  (if (label? x) (labels-add (value *labels*) x (make-lbl))))

(define (err msg)
  (display "*** ") (display msg) (display " *** in: ") (write (value *line*))
  (newline)
  (error "asm: fatal error"))

(define (pass2 x)
  (bind *line* x
    (cond ((label? x)
           (emit-label x))
          ((integer? x)
           (b16 x))
          ((quoted? x)
           (emit-quote (cadr x)))
          ((pair? x)
           (let ((y (emit-16bit 0)))
             (set-car! y (opcode x))))
          (else
           (err "incorrect line")))))

(define (asm l)
  (set! **dynamic-environment** '())
  (if (not (proper-list l)) (error "asm: argument must be a proper list"))
  (bind *labels* (make-labels)
    (begin
      (for-each pass1 l)
      (bind *fixups* (make-fixups)
        (bind *codes* (make-codes)
          (bind *location* 0
            (begin
              (for-each pass2 l)
              (fixup (value *fixups*))
              (codes->list (value *codes*)))))))))

(define (as x)
  (for-each print-hex (asm x)))

; ---------------------------------------------------------------------------



(macro ??
  (lambda (l)
    (define (?-expand l)
      (if (pair? (cdr l))
        `(let ((? ,(car l)))
           (if ? ,(cadr l) ,(?-expand (cddr l))))
        (car l)))
    (?-expand (cdr l))))

(define (couple x)
  (and (pair? x) (pair? (cdr x)) (null? (cddr x))))

(prop-set! 'd0 'd 0)
(prop-set! 'd1 'd 1)
(prop-set! 'd2 'd 2)
(prop-set! 'd3 'd 3)
(prop-set! 'd4 'd 4)
(prop-set! 'd5 'd 5)
(prop-set! 'd6 'd 6)
(prop-set! 'd7 'd 7)
 
(prop-set! 'a0 'a 0)
(prop-set! 'a1 'a 1)
(prop-set! 'a2 'a 2)
(prop-set! 'a3 'a 3)
(prop-set! 'a4 'a 4)
(prop-set! 'a5 'a 5)
(prop-set! 'a6 'a 6)
(prop-set! 'a7 'a 7)
(prop-set! 'sp 'a 7)

(define (b3? x)
  (and (integer? x) (>= x 1) (<= x 8) (lsh (if (= x 8) 0 x) 9)))

(define (b4? x)
  (and (integer? x) (>= x 0) (<= x 16) x))

(define (b8? x)
  (and (integer? x) (>= x -127) (<= x 255) (if (< x 0) (+ x 256) x)))

(define (d? x)
  (prop-ref x 'd))

(define (a? x)
  (prop-ref x 'a))

(define (r? x)
  (?? (d? x) ?
      (a? x) (+ ? 8)
      #!false))

(define (a+? x)
  (and (couple x) (eq? (cadr x) '+) (a (car x))))

(define (a-? x)
  (and (couple x) (eq? (car x) '-) (a (cadr x))))

(define (reg-list? x)
  (and (pair? x) (eq? (car x) 'list)))

(define (label kind x)
  (if (and (couple x) (integer? (car x)))
    (emit-ref kind (labels-get (value *labels*) (cadr x)) (car x))
    (emit-ref kind (labels-get (value *labels*) x) 0)))

(define (b3 x)
  (?? (b3? x) ?
      (err "1 to 8 expected")))

(define (b4 x)
  (?? (b4? x) ?
      (err "0 to 15 expected")))

(define (b8 x)
  (?? (b8? x) ?
      (err "8 bit number expected")))

(define (b16 x)
  (if (and (integer? x) (>= x -32767) (<= x 65535))
    (emit-16bit x)
    (err "16 bit number expected")))

(define (b32 x)
  (if (and (integer? x) (<= x 4294967295)) ; ****** and (>= x -2147483648)
    (begin (emit-16bit (quotient x 65536)) (emit-16bit (- x (* (quotient x 65536) 65536))))
    (err "32 bit number expected")))

(define (b x size)
  (case (logand size  192)
    ((0)  (emit-16bit (b8 x)))
    ((64) (b16 x))
    (else (b32 x))))

(define (d x)
  (?? (d? x) ?
      (err "data register expected")))

(define (a x)
  (?? (a? x) ?
      (err "address register expected")))

(define (a+ x)
  (?? (a+? x) ?
      (err "postincrement addressing mode expected")))

(define (a- x)
  (?? (a-? x) ?
      (err "predecrement addressing mode expected")))

(define (r x)
  (?? (r? x) ?
      (err "register expected")))

(define (reg-list l decr)
  (define (reg-list-aux l decr n)
    (if (atom? l) n
      (if decr (reg-list-aux (cdr l) decr (+ n (lsh 1 (- 15 (r (car l))))))
               (reg-list-aux (cdr l) decr (+ n (lsh 1 (r (car l))))))))
  (reg-list-aux (cdr l) decr 0))

(define (disp x)
  (if (and (couple x) (integer? (car x)) (a? (cadr x)))
    (begin (b16 (car x)) (a (cadr x)))
    (err "address register indirect with displacement mode expected")))

(define (byte op)
  (if (not (= (logand op  192)  0)) (err "size must be byte")))

(define (not-byte op)
  (if (= (logand op  192)  0) (err "size must be word or long")))

(define (word op)
  (if (not (= (logand op  192)  64)) (err "size must be word")))

(define (long op)
  (if (not (= (logand op  192)  128)) (err "size must be long")))

; ---------------------------------------------------------------------------

;
; effective address encoding
;

; addressing mode                      normal syntax    'lisp' syntax
; _______________                      _____________    _____________
;
; data reg direct                      d                d
; adr reg direct                       a                a
; adr reg indirect                     (a)              (a)
; adr reg indirect with postincrement  (a)+             (a +)
; adr reg indirect with predecrement   -(a)             (- a)
; adr reg with displacement            w(a)             (w a)
; adr reg indirect with index          b(a r.W)         (b a r . W)
;                                   or b(a r.L)         (b a r . L)
; absolute short address               w.W              (w . W)
; absolute long address                l.L              (l . L) or ('x)
;                           
; program counter with displacement    p                p
;                                   or n(p)             (n p)
; program counter with index           n(p r.w)         (n p r . w)
;                                   or n(p r.l)         (n p r . l)
; immediate data                       #n               n or 'x
;
; register list                        d0/d4/a1-a3      (list d0 d4 a1 a2 a3)
;
;
; legend:
;              d = d0, d1, ..., d7
;              a = a0, a1, ..., a7, sp
;              r = d or a
;              b = -128 .. 127
;              w = -32768 .. 32767
;              l = -2147483648 .. 2147483647
;              n = integer
;              p = label
;              x = any lisp form


(define (ea x size)     ; effective address encoding
  (define (indirect x y)
    (define (xxx x y z)
      (if (integer? x)
        (cond ((null? z)
               (?? (a? y)      (begin (b16 x) (+ ? 40))
                   (begin (label '16bit (list x y)) 58)))
              (else
               (err "-- ea not implemented yet--")))
        (err "--incorrect ea--")))
    (cond ((pair? y)
           (xxx x (car y) (cdr y)))
          ((null? y)  (?? (a? x)      (+ ? 16)
                          (quoted? x) (begin (emit-quote (cadr x)) 57)
                          (err "--incorrect ea--")))
          ((eq? y 'w)
           (begin (b16 x) 56))
          ((eq? y 'l)
           (begin (b32 x) 57))
          (else
           (err "--incorrect ea--"))))
  (?? (integer? x) (begin (b x size) 60)
      (d? x)       ?
      (a? x)       (+ ? 8)
      (a+? x)      (+ ? 24)
      (a-? x)      (+ ? 32)
      (label? x)   (begin (label '16bit x) 58)
      (quoted? x)  (begin (emit-quote (cadr x)) 60)
      (pair? x)    (indirect (car x) (cdr x))
      (err "--incorrect ea--")))
 
(define (da x size)     ; data addressing mode encoding
  (let ((y (ea x size)))
    (if (or (< y 8) (> y 15)) y
      (err "invalid data addressing mode"))))
 
(define (aa x)          ; alterable addressing mode encoding
  (let ((y (ea x 0)))
    (if (< y 58) y
      (err "invalid alterable addressing mode"))))
 
(define (ca x)          ; control addressing mode encoding
  (let ((y (ea x 0)))
    (if (or (and (> y 15) (< y 24)) (and (> y 39) (< y 60))) y
      (err "invalid control addressing mode"))))
 
(define (caa- x)        ; control alterable or predecrement addressing mode encoding
  (let ((y (ea x 0)))
    (if (or (and (> y 15) (< y 24)) (and (> y 31) (< y 58))) y
      (err "invalid control alterable or predecrement addressing mode"))))
 
(define (ca+ x)         ; control or postincrement addressing mode encoding
  (let ((y (ea x 0)))
    (if (or (and (> y 15) (< y 40)) (and (> y 47) (< y 60))) y
      (err "invalid control or postincrement addressing mode"))))
 
(define (daa x)         ; data alterable addressing mode encoding
  (let ((y (ea x 0)))
    (if (or (< y 8) (and (> y 15) (< y 58))) y
      (err "invalid data alterable addressing mode"))))
 
(define (maa x)         ; memory alterable addressing mode encoding
  (let ((y (ea x 0)))
    (if (and (> y 15) (< y 58)) y
      (err "invalid memory alterable addressing mode"))))

; ---------------------------------------------------------------------------

(define (define-op n l f)
  (let ((descr (cons f n)))
    (for-each (lambda (x) (prop-set! (car x) 'asm (cons (cadr x) descr))) l)))

(define-op -1
  '( (asl.b  -7936) (asl.w  -7872) (asl.l  -7808) 
     (asr.b  -8192) (asr.w  -8128) (asr.l  -8064)
     (lsl.b  -7928) (lsl.w  -7864) (lsl.l  -7800)
     (lsr.b  -8184) (lsr.w  -8120) (lsr.l  -8056)
     (rol.b  -7912) (rol.w  -7848) (rol.l  -7784)
     (ror.b  -8168) (ror.w  -8104) (ror.l  -8040)
     (roxl.b -7920) (roxl.w -7856) (roxl.l -7792) 
     (roxr.b -7920) (roxr.w -7856) (roxr.l -7792) )
  (lambda (op . l)
    (cond ((= (length l) 1)   ; 1 operand
           (word op)
           (+ -8000 (lsh (logand op  24) 6) (logand op  256) (maa (car l))))
          ((= (length l) 2)   ; 2 operands
           (?? (d?  (car l)) (+ op (lsh ? 9) (d (cadr l))  32)
               (b3? (car l)) (+ op ? (d (cadr l)))
               (err "data register or 1..8 expected")))
          (else
           (err "incorrect number of operands")))))

(define-op 0
  '( (illegal  19196)
     (nop      20081)
     (reset    20080)
     (rte      20083)
     (rtr      20087)
     (rts      20085)
     (trapv    20086) )
  (lambda (op)
    op))
  
(define-op 1
  '( (stop  20082) )
  (lambda (op <1>)
    (b16 <1>)
    op))
 
(define-op 1
  '( (trap  20032) )
  (lambda (op <1>)
    (+ op (b4 <1>))))
 
(define-op 1
  '( (ext.w  18560) (ext.l  18624)
     (swap   18496) )
  (lambda (op <1>)
    (+ op (d <1>))))
 
(define-op 1
  '( (unlk  20056) )
  (lambda (op <1>)
    (+ op (a <1>))))

(define-op 1
  '( (bcc  25600)
     (bcs  25856)
     (beq  26368)
     (bge  27648)
     (bgt  28160)
     (bhi  25088)
     (ble  28416)
     (bls  25344)
     (blt  27904)
     (bmi  27392)
     (bne  26112)
     (bpl  27136)
     (bt   24576) (bra  24576)
     (bvc  26624)
     (bvs  26880)
     (bsr  24832) )
  (lambda (op <1>)
    (label 'bra <1>)
    op))
      
(define-op 1
  '( (bra.w  24576) )
  (lambda (op <1>)
    (label '16bit <1>)
    op))
      
(define-op 1
  '( (jmp  20160) (jsr  20096) (pea  18496) )
  (lambda (op <1>)
    (+ op (ca <1>))))
 
(define-op 1
  '( (clr.b   16896) (clr.w   16960) (clr.l   17024)
     (nbcd    18432)
     (neg.b   17408) (neg.w   17472) (neg.l   17536)
     (negx.b  16384) (negx.w  16448) (negx.l  16512)
     (not.b   17920) (not.w   17984) (not.l   18048)
     (scc     21696)
     (scs     21952)
     (seq     22464)
     (sf      20928)
     (sge     23744)
     (sgt     24256)
     (shi     21184)
     (sle     24512)
     (sls     21440)
     (slt     24000)
     (smi     23488)
     (sne     22208)
     (spl     23232)
     (st      20672)
     (svc     22720)
     (svs     22976)
     (tas     19136)
     (tst.b   18944) (tst.w   19008) (tst.l   19072) )
  (lambda (op <1>)
    (+ op (daa <1>))))

(define-op 2
  '( (dbcc  21704)
     (dbcs  21960)
     (dbeq  22472)
     (dbf   20936) (dbra  20936)
     (dbge  23752)
     (dbgt  24264)
     (dbhi  21192)
     (dble  24520)
     (dbls  21448)
     (dblt  24008)
     (dbmi  23496)
     (dbne  22216)
     (dbpl  23240)
     (dbt   20680)
     (dbvc  22728)
     (dbvs  22984) )
  (lambda (op <1> <2>)
    (label '16bit <2>)
    (+ op (d <1>))))

(define-op 2
  '( (bchg  320)
     (bclr  384)
     (bset  448)
     (btst  256) )
  (lambda (op <1> <2>)
    (?? (d?  <1>) (+ op (lsh ? 9) (daa <2>))
        (b8? <1>) (begin (emit-16bit ?) (+  2048 (logand op  192) (daa <2>)))
        (err "data register or -128..255 expected"))))

(define-op 2
  '( (cmpm.b -20216) (cmpm.w -20152) (cmpm.l -20088) )
  (lambda (op <1> <2>)
    (let ((x (a+ <1>)))
      (+ op (lsh (a+ <2>) 9) x))))

(define-op 2
  '( (exg -16128) )
  (lambda (op <1> <2>)
    (let ((r1 (r <1>)))
      (let ((r2 (r <2>)))
        (let ((x (min r1 r2)) (y (max r1 r2)))
          (+ op (lsh (logand x 7) 9) y
            (if (and (<= x 7) (>= y 8))  128  64)))))))

(define-op 2
  '( (abcd   -16128)
     (addx.b -12032) (addx.w -11968) (addx.l -11904)
     (sbcd   -32512)
     (subx.b -28416) (subx.w -28352) (subx.l -28288) )
  (lambda (op <1> <2>)
    (?? (d? <1>)  (+ op (lsh (d <2>) 9) ?)
        (a-? <1>) (+ op (lsh (a- <2>) 9)  8 ?)
        (err "data register or predecrement addressing mode expected"))))

(define-op 2
  '( (add.b -10692) (add.w -10628) (add.l -10564)
     (and.b -15812) (and.w -15748) (and.l -15684)
     (cmp.b -17348) (cmp.w -17284) (cmp.l -17220)
     (eor.b -17860) (eor.w -17796) (eor.l -17732)
     (or.b  -32708) (or.w  -32644) (or.l  -32580)
     (sub.b -27588) (sub.w -27524) (sub.l -27460) )
  (lambda (op <1> <2>)
    (define (stat op)
      (if (not (= (logand op  1024)  0)) (err "invalid destination operand")))
    (define (not-eor op)
      (if (= (logand op  3840)  2560) (err "invalid destination operand")))
    (define (not-stat op)
      (if (= (logand op  1024) 0) (err "invalid destination operand")))
    (define (not-cmp op)
      (if (= (logand op  3840)  3072) (err "invalid source operand")))
    (define (immed-ccr op <1>) (stat op) (byte op)
      (emit-16bit (b8 <1>))
      (logand op  4095))
    (define (immed-sr op <1>) (stat op) (word op)
      (b16 <1>)
      (logand op  4095))
    (define (ea-dreg reg op <1>) (not-eor op)
      (+ (logand op -3904) reg (ea <1> op)))
    (define (ea-areg reg op <1>) (not-stat op) (not-byte op)
      (+ (logand op -4096) reg (lsh (logand op  128) 1)  192 (ea <1> op)))
    (define (dreg-ea reg op <2>) (not-cmp op)
      (+ (logand op -3904) reg  256 (maa <2>)))
    (?? (eq? <2> 'ccr) (immed-ccr op <1>)
        (eq? <2> 'sr ) (immed-sr  op <1>)
        (d? <2>)       (ea-dreg (lsh ? 9) op <1>)
        (a? <2>)       (ea-areg (lsh ? 9) op <1>)
        (d? <1>)       (dreg-ea (lsh ? 9) op <2>)
        (integer? <1>) (begin (b <1> op) (+ (logand op  4032) (daa <2>)))
        (err "invalid operand combination"))))

(define-op 2
  '( (moveq  28672) )
  (lambda (op <1> <2>)
    (+ op (lsh (d <2>) 9) (b8 <1>))))

(define-op 2
  '( (link  20048) )
  (lambda (op <1> <2>)
    (b16 <2>)
    (+ op (a <1>))))
      
(define-op 2
  '( (lea  16832) )
  (lambda (op <1> <2>)
    (+ op (lsh (a <2>) 9) (ca <1>))))
      
(define-op 2
  '( (addq.b  20480) (addq.w  20544) (addq.l  20608)
     (subq.b  20736) (subq.w  20800) (subq.l  20864) )
  (lambda (op <1> <2>)
    (+ op (b3 <1>) (aa <2>))))

(define-op 2
  '( (chk   16768)
     (divs -32320)
     (divu -32576)
     (muls -15936)
     (mulu -16192) )
  (lambda (op <1> <2>)
    (+ op (lsh (d <2>) 9) (da <1> 64))))
 
(define-op 2
  '( (movem.w  18560) (movem.l  18624) )
  (lambda (op <1> <2>)
    (define (emit-movem op l x)
      (set-car! x (reg-list l (= (logand op  56)  32)))
      op)
    (let ((x (emit-16bit 0)))
      (?? (reg-list? <1>) (emit-movem (+ op (caa- <2>)) <1> x)
          (reg-list? <2>) (emit-movem (+ op  1024 (ca+ <1>)) <2> x)
          (err "register list expected")))))

(define-op 2
  '( (movep.w  264) (movep.l  328) )
  (lambda (op <1> <2>)
    (?? (d? <1>) (+ op  128 (lsh ? 9) (disp <2>))
        (d? <2>) (+ op (lsh ? 9) (disp <1>))
        (err "invalid operands"))))

(define-op 2
  '( (move.b   4096) (move.w   12352) (move.l   8320) )
  (lambda (op <1> <2>)
    (cond ((eq? <1> 'sr ) (word op) (+  16576 (daa <2>)))
          ((eq? <1> 'usp) (long op) (+  20072 (a <2>)))
          ((eq? <2> 'ccr) (word op) (+  17600 (da <1> op)))
          ((eq? <2> 'sr ) (word op) (+  18112 (da <1> op)))
          ((eq? <2> 'usp) (long op) (+  20064 (a <1>)))
          (else
           (let ((x (ea <1> op)))
             (let ((y (aa <2>)))
               (and (<= 8 y) (<= y 15) (not-byte op))
               (+ (logand op -4096) (logand (+ (lsh y 3) (lsh y 9))  4032) x)))))))


; ---------------------------------------------------------------------------

(define (ass x)
  (let ((l (tlist)))
    (define (convert y)
      (if (integer? y)
        (begin
          (if (negative? y) (set! y (+ y 65536)))
          (tlist-add l (quotient y 256))
          (tlist-add l (remainder y 256)))
        (let ((z (+ (* (typetag (cadr y)) 16777216)
                    (typetag-set! (cadr y) 0))))
          (tlist-add l (quotient z 16777216))
          (tlist-add l (remainder (quotient z 65536) 256))
          (tlist-add l (remainder (quotient z 256) 256))
          (tlist-add l (remainder z 256)))))
    (for-each convert (asm x))
    (tlist->list l)))

(define (assemble x)
  (set-code! (eval '(lambda () 0))
             (list->bytevector
               (ass (append
                      (append '(32610
                                (movem.l (list d0 d1 d2 d3 d4 d5 d6
                                               a0 a1 a2 a3 a5 a6) (- a7))
                                (move.b d7 d0) (asl.l 2 d7)
                                (move.l a2 a0) (add.l d7 a0)
                                (move.l (- a0) a1) (move.l (- a0) a2)
                                (move.l (- a0) a3) (move.l (- a0) a4)
                                (move.l (- a0) a5) (move.l (- a0) a6)
                                (lea $return1$ a0))
                              x)
                      '($return1$
                        (move.l a1 d7)
                        (movem.l (a7 +) (list d0 d1 d2 d3 d4 d5 d6
                                              a0 a1 a2 a3 a5 a6))
                        (lea $return2$ a4)
                        (jmp (a6))
                        $return2$
                        -26624))))))
