;------------------------------------------------------------------------------
;
; The file compiler:
; -----------------

(define (compile-file source . optional)

  (let* ((source-port (open-input-file (string-append source ".scm")))
         (target-port (open-output-file (string-append source ".s")))
         (asm-output? (not (null? optional)))
         (asm-port (and asm-output?
                        (if (string? (car optional))
                          (open-output-port (car optional))
                          (current-output-port)))))

  (define (parse expression env)
    (if (define-expr? expression)
      (let ((name (definition-variable expression)))
        (newline) (display "parsing define to ") (write name)
        (list 'define (lookup-var name env)
          (expression->parse-tree (definition-value expression) env)))
      (begin
        (newline) (display "parsing an expression")
        (list 'eval (expression->parse-tree expression env)))))

  (define (read-program port env)
    (let ((expression (read port)))
      (if (eof-object? expression)
        '()
        (let ((x (parse expression env)))
          (cons x (read-program port env))))))

  (define (dump-objects)
    (write-code "|--------------------------------")
    (do ((l (list-tail (queue->list object-queue)
                       (length initial-object-queue))
            (after-first l)))
        ((null? l))
      (write-code)
      (let ((x (first l)))
        (let ((obj (car x)) (num (cdr x)))
          (cond ((pair? obj)
                 (write-code "PAIR_OBJECT(obj" num ","
                             (encode-object (car obj)) ","
                             (encode-object (cdr obj)) ")"))
                ((symbol? obj)
                 (write-code "SYMBOL_OBJECT(obj" num ","
                             (encode-object (symbol->string obj)) ")"))
                ((string? obj)
                 (write-code "STRING_OBJECT(obj" num ")")
                 (for-each
                   (lambda (l) (write-code tab ".byte" tab (list-of l ",")))
                   (partition (map char->integer (string->list obj)) 10))
                 (write-code "END"))
                ((vector? obj)
                 (write-code "VECTOR_OBJECT(obj" num ")")
                 (for-each
                   (lambda (l) (write-code tab ".long" tab (list-of l ",")))
                   (partition (map encode-object (vector->list obj)) 10))
                 (write-code "END"))
                (else
                 (error "unknown object kind"))))))
    (write-code)
    (write-code "PAIR_OBJECT(symbol_table,symbol_table1,"
                (encode-object symbol-table)
                ")")
    (write-code "SYMBOL_OBJECT(symbol_table1,symbol_table2)")
    (write-code "STRING_OBJECT(symbol_table2)")
    (write-code "END"))

  (set! *global-env* (make-standard-environment))
  (let ((infile (string-append filename ".scm"))
        (outfile (string-append filename ".s")))
    (let ((input (open-input-file infile)))
      (set! *code-port* (open-output-file outfile))
      (let ((program (read-program input *global-env*)))
        (set! object-counter (make-counter initial-object-counter))
        (set! object-queue (list->queue (append initial-object-queue '())))
        (set! symbol-table initial-symbol-table)
        (newline)
        (compile-procedures program)
        (compile-top-level program)
        (newline)
        (newline) (display "dumping objects")
        (dump-objects)
        (newline))
      (close-input-port input)
      (close-output-port *code-port*))))

(define (encode-object obj)
  (let ((obj-repr (object-repr obj)))
    (if obj-repr
      obj-repr
      (list "obj" (add-object obj)))))

(define *global-env* '())

(define (make-standard-environment)
  (let ((env (make-empty-environment)))
    (set! symbol-table-var (lookup-var 1 env))
    (for-each (lambda (name) (lookup-var name env))
              (map first *primitive-procedures*))
    env))

;------------------------------------------------------------------------------
;
; Runtime context manipulation (i.e. where the variables are, what registers
; are in use, etc.)

; runtime context description: frame-size = number of slots presently allocated
; for the current frame on the stack, slots = list of variables associated with
; each slot (topmost slot first), regs = list of variables contained in each
; register, closed = list of variables which are closed with respect to the
; current procedure.

(define-structure context frame-size slots regs closed)

(define frame-size '())
(define slots      '())
(define regs       '())
(define closed     '())

; special variable used to denote the continuation of a procedure

(define cont-var (make-var 'cont #t set-empty set-empty))
(define cont-var-set (set-singleton cont-var))

; special variable used to denote the pointer to the closed variables

(define closed-var (make-var 'closed #t set-empty set-empty))

; (entry-context proc closed) returns the context in existence upon entry to
; the procedure `proc'.

(define (entry-context proc closed)
  (let ((rest (proc-rest proc))
        (parms (proc-parms proc)))

    (define (locate-parms parms slots)
      (if (<= (length parms) (if rest 1 nb-regs))
        (if (null? closed)
          (make-context
            (length slots)
            slots
            (cons cont-var parms)
            closed)
          (make-context
            (+ (length slots) 1)
            (cons closed-var slots)
            (cons cont-var parms)
            closed))
        (locate-parms (cdr parms) (cons (car parms) slots))))

    (locate-parms parms '())))

(define (restore-context context)
  (set! frame-size (context-frame-size context))
  (set! slots      (context-slots context))
  (set! regs       (context-regs context))
  (set! closed     (context-closed context)))

(define (current-context)
  (make-context frame-size slots regs closed))

(define (flush-regs)
  (set! regs '()))

(define (get-var opnd)
  (cond ((glob? opnd)
         (glob-var opnd))
        ((reg? opnd)
         (list-ref regs (reg-num opnd)))
        ((slot? opnd)
         (list-ref slots (- frame-size (slot-num opnd))))
        (else
         (error "location must be global, register or stack slot"))))

(define (put-var opnd new)
  (cond ((reg? opnd)
         (set! regs (replace-nth regs (reg-num opnd) new)))
        ((slot? opnd)
         (set! slots (replace-nth slots (- frame-size (slot-num opnd)) new)))
        (else
         (error "location must be register or stack slot"))))

(define (alloc-slots n)
  (set! frame-size (+ frame-size n))
  (let loop ((i n))
    (if (> i 0)
      (begin
        (set! slots (cons (make-arg-var '()) slots))
        (loop (- i 1))))))

(define (dealloc-slots n)
  (set! frame-size (- frame-size n))
  (set! slots      (list-tail slots n)))

(define (push-slot)
  (alloc-slots 1))

(define (pop-slot)
  (dealloc-slots 1))

(define (replace-nth l i v)
  (if (null? l)
    (if (= i 0)
      (list v)
      (cons (make-arg-var '()) (replace-nth l (- i 1) v)))
    (if (= i 0)
      (cons v (cdr l))
      (cons (car l) (replace-nth (cdr l) (- i 1) v)))))

(define (pos-in-list x l)
  (let ((y (memq x l)))
    (and y (- (length l) (length y)))))

(define (live-vars live)
  (if (not (set-empty? (set-intersection live (list->set closed))))
    (set-adjoin live closed-var)
    live))

(define (dead-slots live)
  (let ((live-v (live-vars live)))
    (define (loop s l i)
      (cond ((null? l) (list->set (reverse s)))
            ((set-member? (first l) live-v)
             (loop s (after-first l) (- i 1)))
            (else
             (loop (cons i s) (after-first l) (- i 1)))))
    (loop '() slots frame-size)))

(define (live-slots live)
  (let ((live-v (live-vars live)))
    (define (loop s l i)
      (cond ((null? l) (list->set (reverse s)))
            ((set-member? (first l) live-v)
             (loop (cons i s) (after-first l) (- i 1)))
            (else
             (loop s (after-first l) (- i 1)))))
    (loop '() slots frame-size)))

(define (dead-regs live)
  (let ((live-v (live-vars live)))
    (define (loop s l i)
      (cond ((> i nb-regs) (list->set (reverse s)))
            ((null? l)
             (loop (cons i s) l (+ i 1)))
            ((and (set-member? (first l) live-v)
                  (not (memq (first l) slots)))
             (loop s (after-first l) (+ i 1)))
            (else
             (loop (cons i s) (after-first l) (+ i 1)))))
    (loop '() regs 0)))

(define (live-regs live)
  (let ((live-v (live-vars live)))
    (define (loop s l i)
      (cond ((null? l) (list->set (reverse s)))
            ((and (set-member? (first l) live-v)
                  (not (memq (first l) slots)))
             (loop (cons i s) (after-first l) (+ i 1)))
            (else
             (loop s (after-first l) (+ i 1)))))
    (loop '() regs 0)))

(define (lowest-dead-slot live)
  (make-slot (or (lowest (dead-slots live)) (+ frame-size 1))))

(define (highest-live-slot live)
  (make-slot (or (highest (live-slots live)) 0)))

(define (lowest-dead-reg live)
  (let ((x (lowest (set-remove (dead-regs live) 0))))
    (if x (make-reg x) #f)))

(define (highest-dead-reg live)
  (let ((x (highest (dead-regs live))))
    (if x (make-reg x) #f)))

(define (highest set) ; return highest number in the set
  (if (set-empty? set) #f (apply max (set->list set))))

(define (lowest set) ; return lowest number in the set
  (if (set-empty? set) #f (apply min (set->list set))))

(define (above set n) ; return numbers above n in the set
  (set-keep (lambda (x) (> x n)) set))
  
(define (below set n) ; return numbers below n in the set
  (set-keep (lambda (x) (< x n)) set))
  
(define (var->opnd* var)
  (if (global? var)
    (make-glob var)
    (let ((n (pos-in-list var regs)))
      (if n
        (make-reg n)
        (let ((n (pos-in-list var slots)))
          (if n
            (make-slot (- frame-size n))
            (let ((n (pos-in-list var closed)))
              (if n
                (make-indx (var->opnd* closed-var)
                           (+ first-closed-offset (* object-size n)))
                (let ((x (assq var constant-vars)))
                  (if x
                    (cdr x)
                    (error "variable is not accessible")))))))))))

(define (var->opnd var)
  (if (and (not (global? var)) (mutable? var))
    (make-indx (var->opnd* var) cell-offset)
    (var->opnd* var)))

;------------------------------------------------------------------------------

(define-structure counter value)

(define instr-queue '())
(define proc-queue '())
(define const-queue '())
(define label-counter '())
(define object-counter '())
(define object-queue '())
(define constant-vars '())
(define known-procs '())
(define symbol-table '())
(define symbol-table-var '())

;------------------------------------------------------------------------------

(define (compile-top-level program)

  (define (compile-list l)
    (if (not (null? l))
      (begin
        (cond ((eq? (caar l) 'define)
               (let* ((var (cadar l))
                      (name (var-name var))
                      (value (caddar l)))
                 (newline) (display "compiling define to ") (write name)
                 (let ((opnd (if (proc? value)
                               (make-prim name #f)
                               (gen-node value cont-var-set 'need))))
                   (emit-copy opnd (var->opnd var) cont-var-set))))
              ((eq? (caar l) 'eval)
               (let ((value (cadar l)))
                 (newline) (display "compiling an expression")
                 (gen-node value cont-var-set
                   (if (null? (after-first l)) 'tail 'side)))))
        (compile-list (cdr l)))))

  (newline)

  (set! instr-queue (queue-empty))
  (set! proc-queue (queue-empty))
  (set! const-queue (queue-empty))
  (queue-put! const-queue symbol-table-var) ; space for symbol table pointer
  (set! label-counter (make-counter 0))
  (set! constant-vars '())
  (set! known-procs '())

  (restore-context (make-context 0 '() (list cont-var) '()))

  (compile-list program)

  (do () ((queue-empty? proc-queue))
    (apply gen-proc (queue-get! proc-queue)))

  (write-code "|--------------------------------")
  (write-code)
  (write-code "DEFINE_PRIMITIVE(`program-top-level')")

  (do ((l (cons (list 'copy (make-imm symbol-table-var)
                            (var->opnd symbol-table-var))
                (queue->list instr-queue))
          (after-first l)))
      ((null? l))
    (write-instr (first l)))

  (write-code "CONST(symbol_table)")

  (do ((l (after-first (queue->list const-queue)) (after-first l)))
      ((null? l))
    (write-code "CONST(obj" (add-object (first l)) ")"))

  (write-code "END")
  (write-code))

;------------------------------------------------------------------------------

(define (compile-procedures l)
  (if (not (null? l))
    (begin
      (if (eq? (caar l) 'define)
        (let* ((var (cadar l))
               (name (var-name var))
               (value (caddar l)))
          (if (proc? value)
            (begin
              (newline) (display "compiling procedure ") (write name)
              (compile-procedure name value)))))
    (compile-procedures (cdr l)))))

;------------------------------------------------------------------------------

(define (compile-procedure name proc)

  (set! instr-queue (queue-empty))
  (set! proc-queue (queue-empty))
  (set! const-queue (queue-empty))
  (set! label-counter (make-counter 0))
  (set! constant-vars '())
  (set! known-procs '())

  (if (not compile-primitives) (emit-check-stack))

  (gen-body proc '())
  (do () ((queue-empty? proc-queue))
    (apply gen-proc (queue-get! proc-queue)))

  (write-code (if compile-primitives "DEFINE_PRIMITIVE" "DEFINE_PROCEDURE")
    "(`" name "'," (length (proc-parms proc))
    (if (proc-rest proc) ",REST" '())
    ")")

  (do ((l (queue->list instr-queue) (after-first l)))
      ((null? l))
    (write-instr (first l)))

  (do ((l (queue->list const-queue) (after-first l)))
      ((null? l))
    (write-code "CONST(obj" (add-object (first l)) ")"))

  (write-code "END")
  (write-code))

(define (new-label)
  (counter-value-set! label-counter (+ (counter-value label-counter) 1))
  (counter-value label-counter))

(define (add-constant val)
  (let ((n (pos-in-list val (queue->list const-queue))))
    (if n
      (+ n 1)
      (begin      
        (queue-put! const-queue val)
        (queue-length const-queue)))))

(define (add-object obj)
  (let ((x (assq obj (queue->list object-queue))))
    (if x
      (cdr x)
      (begin
        (if (and (symbol? obj) (not (memq obj symbol-table)))
          (begin
            (set! symbol-table (cons obj symbol-table))
          (counter-value-set! object-counter
            (+ (counter-value object-counter) 1))
          (queue-put! object-queue
            (cons symbol-table (counter-value object-counter)))))
        (counter-value-set! object-counter
          (+ (counter-value object-counter) 1))
        (queue-put! object-queue
          (cons obj (counter-value object-counter)))
        (counter-value object-counter)))))

(define (add-constant-var var opnd)
  (set! constant-vars (cons (cons var opnd) constant-vars)))

(define (constant-var? var)
  (assq var constant-vars))

(define (not-constant-var? var)
  (not (constant-var? var)))

(define (add-known-proc label proc)
  (set! known-procs (cons (cons label proc) known-procs)))

;------------------------------------------------------------------------------

; generate code for a procedure

(define (gen-proc proc label closed)
  (emit-entry-label label)
  (emit-enter (proc-rest proc) (not (null? closed)) (length (proc-parms proc)))
  (emit-label (+ label 1))
  (emit-check-stack)
  (gen-body proc closed))

(define (schedule-gen-proc proc closed-list)
  (let* ((label1 (new-label))  ; arg check entry point
         (label2 (new-label))) ; no arg check entry point
    (queue-put! proc-queue (list proc label1 closed-list))
    (make-lbl label1)))
 
(define (prepare-closure proc-var proc-label closed-list live)
  (emit-copy proc-label
             (make-indx (var->opnd proc-var) proc-offset)
             (set-union (list->set closed-list) (set-adjoin live proc-var)))
  (let loop ((l closed-list) (i 0))
    (if (not (null? l))
      (begin
        (emit-copy (var->opnd* (first l))
                   (make-indx (var->opnd proc-var)
                              (+ first-closed-offset (* object-size i)))
                   (set-union (list->set (after-first l))
                              (set-adjoin live proc-var)))
        (loop (after-first l) (+ i 1))))))

;------------------------------------------------------------------------------

; generate code for the body of a procedure

(define (gen-body proc closed)
  (let ((live (set-adjoin (free-variables (proc-body proc)) cont-var)))
    (restore-context (entry-context proc closed))
    (for-each
      (lambda (parm)
        (if (mutable? parm) (emit-make-cell (var->opnd* parm) live)))
      (proc-parms proc))
    (gen-node (proc-body proc) cont-var-set 'tail)))

;------------------------------------------------------------------------------

; generate code for an expression

(define (gen-node node live why)

  (cond ((root? node)
         (error "Root nodes should not appear inside parse trees"))

        ((uns? node)
         (gen-return result-reg why))

        ((cst? node)
         (gen-return
           (let ((x (primitive-procedure->name (cst-val node))))
             (if x (make-prim x #f) (make-imm (cst-val node))))
           why))

        ((ref? node)
         (gen-return
           (cond ((eq? why 'side)
                  result-reg)
                 ((and assume-standard-prims
                       (global? (ref-var node))
                       (primitive-procedure (var-name (ref-var node))))
                  (make-prim (var-name (ref-var node)) #f))
                 (else
                  (var->opnd (ref-var node))))
           why))

        ((set? node)
         (let* ((src (gen-node (set-val node)
                               (set-adjoin live (set-var node))
                               'keep))
                (dst (var->opnd (set-var node))))
           (emit-copy src dst live))
         (gen-return result-reg why))

        ((cond? node)
         (gen-cond (cond-pre node) (cond-con node) (cond-alt node) live why))

        ((proc? node)
         (let* ((closed (not-constant-closed-vars node))
                (closed-list (set->list closed))
                (proc-label (schedule-gen-proc node closed-list)))
           (let ((opnd
                  (if (null? closed-list)
                    (begin
                      (add-known-proc (lbl-num proc-label) node)
                      proc-label)
                    (begin
                      (emit-dealloc (- frame-size
                                       (slot-num (highest-live-slot
                                                   (set-union closed live)))))
                      (emit-make-closures (list (length closed-list)))
                      (let ((slot (make-slot frame-size))
                            (var (make-arg-var 'closure)))
                        (put-var slot var)
                        (prepare-closure var proc-label closed-list live)
                        slot)))))
             (gen-return opnd why))))

        ((call? node)
         (gen-call node live why))

        (else
         (error "Unknown parse tree node type"))))

(define (gen-return opnd why)
  (cond ((eq? why 'tail)
         (emit-copy opnd result-reg cont-var-set)
         (emit-jump #f 0 #f (var->opnd cont-var) set-empty))
        (else
         opnd)))

(define (not-constant-closed-vars proc)
  (set->list (set-keep not-constant-var? (closed-variables proc))))

;------------------------------------------------------------------------------

(define (gen-cond pre con alt live why)

  (define (gen-test not-dir test args)

    (if (not (eq? why 'tail)) ; could be removed if merge-contexts was smarter.
      ; save regs on stack if they contain values needed after the conditional
      (save-regs (live-regs live)
                 (set-union live
                            (apply set-union (map free-variables args))
                            (free-variables con)
                            (free-variables alt))))

    (let loop ((args* args)
               (liv (set-union live (free-variables con) (free-variables alt)))
               (vars* '()))
      (if (not (null? args*))

        (let* ((needed (vals-live-vars liv (after-first args*)))
               (var
                (save-var (gen-node (first args*) needed 'need)
                          (make-arg-var 'predicate)
                          needed)))
          (loop (after-first args*) (set-adjoin liv var) (cons var vars*)))

        (let ((label1 (new-label)))

          (emit-if not-dir test (map var->opnd (reverse vars*)) label1
            (set-union live (free-variables con) (free-variables alt)))

          (let ((pre-context (current-context)))
            (if (eq? why 'tail)

              (begin
                (gen-node con live why)
                (restore-context pre-context)
                (emit-label label1)
                (gen-node alt live why))

              (begin

                (let ((con-opnd (gen-node con live why)))
                  (if (and (set-member? 1 (live-regs live))
                           (not (and (reg? con-opnd)
                                     (= (reg-num con-opnd) 1))))
                    (save-opnd result-reg live))
                  (emit-copy con-opnd result-reg live)
                  (put-var result-reg (make-arg-var 'consequent)))

                (let ((con-instr-queue instr-queue)
                      (con-context (current-context)))
                  (set! instr-queue (queue-empty))
                  (restore-context pre-context)

                  (let ((alt-opnd (gen-node alt live why)))
                    (if (and (set-member? 1 (live-regs live))
                             (not (and (reg? alt-opnd)
                                       (= (reg-num alt-opnd) 1))))
                      (save-opnd result-reg live))
                    (emit-copy alt-opnd result-reg live)
                    (put-var result-reg (make-arg-var 'alternative)))

                  (let ((alt-instr-queue instr-queue)
                        (alt-context (current-context)))
                    (set! instr-queue con-instr-queue)
                    (restore-context con-context)

                    (merge-contexts alt-context alt-instr-queue live)

                    (if (not (queue-empty? alt-instr-queue))
                      (let ((label2 (new-label)))
                        (emit-jump #f frame-size #f (make-lbl label2) live)
                        (emit-label label1)
                        (for-each put-instr (queue->list alt-instr-queue))
                        (emit-label label2))
                      (emit-label label1))

                    (put-var result-reg (make-arg-var 'result))
                    result-reg)))))))))

  (define (gen-cond* dir pre)
    (let ((known-prim
           (and assume-standard-prims
                (call? pre)
                (ref? (call-oper pre))
                (global? (ref-var (call-oper pre)))
                (assq (var-name (ref-var (call-oper pre)))
                      '( (not 1 #f) (null? 1 #f) (pair? 1 #f)
                         (char? 1 #f) (procedure? 1 #f) (string? 1 #f)
                         (vector? 1 #f) (symbol? 1 #f) (eq? 2 #f)
                         (eqv? 2 #f) (= 2 #t) (char=? 2 #t) (< 2 #t)
                         (char<? 2 #t) (> 2 #t) (char>? 2 #t)
                         (<= 2 #t) (char<=? 2 #t) (>= 2 #t)
                         (char>=? 2 #t) )))))
      (if (and known-prim
               (= (length (call-args pre)) (second known-prim))
               (or (not (third known-prim)) assume-types-ok))
        (if (eq? (first known-prim) 'not)
          (gen-cond* (not dir) (first (call-args pre)))
          (gen-test (eq? dir (not (memq (first known-prim)
                                        '(<= char<=? >= char>=?))))
                    (cond ((memq (first known-prim) '(eq? eqv? = char=?)) '=)
                          ((memq (first known-prim) '(< char<? >= char>=?)) '<)
                          ((memq (first known-prim) '(> char>? <= char<=?)) '>)
                          (else (first known-prim)))
                    (call-args pre)))
        (gen-test dir 'true? (list pre)))))

  (gen-cond* #f pre))

(define (merge-contexts alt-context alt-instr-queue live)
  (let ((live-v (live-vars live)))
    (if (or (not (same-var-list?
                   (reverse (context-slots alt-context))
                   (reverse slots)
                   live-v))
            (not (same-var-list?
                   (context-regs alt-context)
                   regs
                   (set-difference live-v (list->set slots)))))
      (error "context merging for conditionals not yet implemented")
      (let ((dealloc-to (slot-num (highest-live-slot live))))
        (emit-dealloc (- frame-size dealloc-to))
        (let ((con-instr-queue instr-queue)
              (con-context (current-context)))
          (set! instr-queue alt-instr-queue)
          (restore-context alt-context)
          (emit-dealloc (- frame-size dealloc-to))
          (set! instr-queue con-instr-queue)
          (restore-context con-context)
          (set! regs (cancel-diff-vars regs (context-regs alt-context))))))))

(define (same-var-list? l1 l2 live-v)
  (cond ((null? l1)
         (or (null? l2) (same-var-list? l2 l1 live-v)))
        ((null? l2)
         (and (not (set-member? (first l1) live-v))
              (same-var-list? (after-first l1) '() live-v)))
        ((eq? (first l1) (first l2))
         (same-var-list? (after-first l1) (after-first l2) live-v))
        (else
         (and (not (set-member? (first l1) live-v))
              (not (set-member? (first l2) live-v))))))

(define (cancel-diff-vars l1 l2)
  (cond ((null? l1)
         '())
        ((null? l2)
         '())
        ((eq? (first l1) (first l2))
         (cons (first l1)
               (cancel-diff-vars (after-first l1) (after-first l2))))
        (else
         (cons '()
               (cancel-diff-vars (after-first l1) (after-first l2))))))

;------------------------------------------------------------------------------

; generate code for a procedure call

(define (gen-call node live why)
  (let* ((oper (call-oper node))
         (args (call-args node))
         (nb-args (length args)))

    (if (and (proc? oper) ; applying a lambda-expr is like a 'let' or 'letrec'
             (not (proc-rest oper))
             (= (length (proc-parms oper)) nb-args))

      (gen-let (proc-parms oper) args (proc-body oper) live why)

      (if (open-codable? node)

        (let loop ((args* args) (liv live) (vars* '()))
          (if (not (null? args*))

            (let* ((needed (vals-live-vars liv (after-first args*)))
                   (var
                    (save-var (gen-node (first args*) needed 'need)
                              (make-arg-var 'arg)
                              needed)))
              (loop (after-first args*) (set-adjoin liv var) (cons var vars*)))

            (let ((result-var (make-arg-var 'result))
                  (result (or (lowest-dead-reg live) (lowest-dead-slot live))))

              (emit-open-code (make-prim (var-name (ref-var oper)) nb-args)
                              (map var->opnd (reverse vars*))
                              result
                              live)

              (put-var result result-var)

              (gen-return result why))))

      (let ((eval-order (arg-eval-order oper args))
            (live-after (if (eq? why 'tail) (set-remove live cont-var) live)))

        ; save regs on stack if they contain values needed after the call
        (save-regs (live-regs live-after) (args-live-vars live eval-order))

        (let ((frame-start (slot-num (highest-live-slot live-after))))

          (let loop ((l eval-order) (x live) (stacked '()) (oper-var '()))
            (if (null? l)

              (let* ((oper-opnd (var->opnd oper-var))
                     (exist-entry-point
                       (or (and (prim? oper-opnd)
                                (primitive-entry-point (prim-name oper-opnd)
                                                       nb-args))
                           (and (lbl? oper-opnd)
                                (let ((x (assq (lbl-num oper-opnd) known-procs)))
                                  (and x
                                       (let ((proc (cdr x)))
                                         (and (not (proc-rest proc))
                                              (= (length (proc-parms proc))
                                                 nb-args))))))))
                     (nb-args-check
                       (if exist-entry-point #f nb-args))
                     (opnd
                       (cond ((and exist-entry-point (prim? oper-opnd))
                              (make-prim (prim-name oper-opnd) nb-args))
                             ((and exist-entry-point (lbl? oper-opnd))
                              (make-lbl (+ (lbl-num oper-opnd) 1)))
                             (else
                              oper-opnd))))

                (do ((p stacked (after-first p)))
                    ((null? p))
                  (let ((slot (first (first p)))
                        (reg (second (first p)))
                        (liv (third (first p)))
                        (var (fourth (first p))))
                    (emit-copy slot reg liv)
                    (put-var reg var)))

                (if (eq? why 'tail)
                  (let ((liv (set-remove x cont-var)))
                    (emit-copy (var->opnd cont-var) cont-reg liv)
                    (emit-jump (not (or (prim? opnd) (lbl? opnd)))
                      (+ frame-start (max 0 (- nb-args nb-regs)))
                      nb-args-check
                      opnd
                      (set-union (set-remove liv oper-var) live-after)))
                  (emit-call (not (or (prim? opnd) (lbl? opnd)))
                    (+ frame-start (max 0 (- nb-args nb-regs)))
                    nb-args-check
                    opnd
                    (set-union (set-remove x oper-var) live-after)))

                (dealloc-slots (- frame-size frame-start))
                (flush-regs)
                (put-var result-reg (make-arg-var 'result))
                result-reg)

              (let* ((arg (car (first l)))
                     (pos (cdr (first l)))
                     (needed (args-live-vars x (after-first l)))
                     (opnd (gen-node arg needed 'need))
                     (var (make-arg-var pos)))
                (if (= pos 0) ; operator
                  (if (and (ref? arg)
                           (not (or (imm? opnd) (prim? opnd) (lbl? opnd))))
                    (loop (after-first l)
                          (set-adjoin x (ref-var arg))
                          stacked
                          (ref-var arg))
                    (begin
                      (save-arg opnd var needed)
                      (loop (after-first l)
                            (set-adjoin x var)
                            stacked
                            var)))
                    (if (<= pos (- nb-args nb-regs)) ; arg passed on stack
                       (let ((slot (make-slot (+ frame-start pos))))
                         (if (set-member? (slot-num slot) (live-slots needed))
                           (save-opnd slot
                             (if (slot? opnd)
                               (set-adjoin needed (get-var opnd))
                               needed)))
                         (emit-copy opnd slot needed)
                         (put-var slot var)
                         (loop (after-first l)
                               (set-adjoin x var)
                               stacked
                               oper-var))
                       (let ((reg (make-reg (- pos (max 0 (- nb-args nb-regs))))))
                         (if (every-trivial? (after-first l))
                           (begin
                             (if (set-member? (reg-num reg) (live-regs needed))
                               (save-opnd reg
                                 (if (slot? opnd)
                                   (set-adjoin needed (get-var opnd))
                                   needed)))
                             (emit-copy opnd reg needed)
                             (put-var reg var)
                             (loop (after-first l)
                                   (set-adjoin x var)
                                   stacked
                                   oper-var))
                           (let ((slot (make-slot (+ frame-size 1))))
                             (emit-copy opnd slot needed)
                             (put-var slot var)
                             (loop (after-first l)
                                   (set-adjoin x var)
                                   (cons (list slot reg x var) stacked)
                                   oper-var)))))))))))))))

(define (save-opnd opnd live)
  (let ((slot (lowest-dead-slot live)))
    (emit-copy opnd slot live)
    (put-var slot (get-var opnd))))

(define (save-regs regs live)
  (for-each (lambda (i) (save-opnd (make-reg i) live)) (set->list regs)))

(define (every-trivial? l)
  (or (null? l)
      (and (trivial? (car (first l)))
           (every-trivial? (after-first l)))))

(define (trivial? node)
  (or (uns? node)
      (cst? node)
      (ref? node)
      (and (set? node) (trivial? (set-val node)))))

(define (open-codable? node)
  (and assume-standard-prims
       assume-types-ok
       (call? node)
       (ref? (call-oper node))
       (global? (ref-var (call-oper node)))
       (let ((x (assq (var-name (ref-var (call-oper node)))
                      '( (+ 2) (- 2) (* 2) (/ 2)
                         (car 1) (cdr 1) (set-car! 2) (set-cdr! 2) ))))
         (and x
              (= (length (call-args node)) (second x))
              (or assume-fixnum-arith
                  (not (memq (first x) '(+ - * /))))))))

(define (make-arg-var pos)
  (make-var pos #t set-empty set-empty))

(define (arg-eval-order oper args)
  (let ((nb-args (length args)))

    (define (loop1 args pos first-args)
      (if (and (not (null? args)) (<= pos (- nb-args nb-regs)))
        (loop1 (after-first args)
               (+ pos 1)
               (cons (cons (first args) pos) first-args))
        (append (reverse first-args) (loop2 args pos '() '()))))

    (define (loop2 args pos part1 part2)
      (cond ((null? args)
             (if (trivial? oper)
                (append part1 part2 (list (cons oper 0)))
                (append (cons (cons oper 0) part1) part2)))
            ((trivial? (first args))
             (loop2 (after-first args)
                    (+ pos 1)
                    part1
                    (cons (cons (first args) pos) part2)))
            (else
             (loop2 (after-first args)
                    (+ pos 1)
                    (cons (cons (first args) pos) part1)
                    part2))))

    (loop1 args 1 '())))

(define (args-live-vars live order)
  (if (null? order)
    live
    (args-live-vars (set-union live (free-variables (car (first order))))
                    (after-first order))))

;------------------------------------------------------------------------------

; generate code for a 'let' or 'letrec'

(define (gen-let vars vals node live why)
  (let ((var-val-map (map cons vars vals))
        (var-set (list->set vars))
        (all-live (apply set-union live
                    (free-variables node)
                    (map free-variables vals))))

    (define (var->val var) (cdr (assq var var-val-map)))

    (define (proc-var? var) (proc? (var->val var)))

    (define (non-mutable? var) (not (mutable? var)))

    (let ((proc-vars (set-keep proc-var? var-set)))

      (let loop1 ((const-proc-vars (set-keep non-mutable? proc-vars)))

        (define (closed-vars var)
          (set-difference (not-constant-closed-vars (var->val var))
                          const-proc-vars))

        (define (no-closed-vars? var)
          (set-empty? (closed-vars var)))

        (define (closed-vars? var)
          (not (no-closed-vars? var)))

        (let ((new-const-proc-vars (set-keep no-closed-vars? const-proc-vars)))
          (if (not (set-equal? new-const-proc-vars const-proc-vars))
            (loop1 new-const-proc-vars)

            (let ((clo-vars (set-keep closed-vars? proc-vars)))

              (for-each
                (lambda (proc-var)
                  (let ((label (schedule-gen-proc (var->val proc-var) '())))
                    (add-known-proc (lbl-num label) (var->val proc-var))
                    (add-constant-var proc-var label)))
                (set->list const-proc-vars))

              (if (not (set-empty? clo-vars))
                (let ((clo-vars-list (set->list clo-vars)))
                  (emit-dealloc
                    (- frame-size (slot-num (highest-live-slot all-live))))
                  (emit-make-closures (map length (map closed-vars clo-vars-list)))
                  (let loop2 ((l clo-vars-list)
                              (i (+ (- frame-size (length clo-vars-list)) 1)))
                    (if (not (null? l))
                      (let ((var (first l))
                            (slot (make-slot i)))
                        (put-var slot var)
                        (if (mutable? var) (emit-make-cell slot all-live))
                        (loop2 (after-first l) (+ i 1)))))
                  (for-each
                    (lambda (proc-var)
                      (let ((closed-list (set->list (closed-vars proc-var))))
                        (prepare-closure proc-var
                                         (schedule-gen-proc (var->val proc-var)
                                                            closed-list)
                                         closed-list
                                         all-live)))
                    clo-vars-list)))

              (let ((other-vars-list
                      (set->list
                        (set-keep (lambda (var)
                                    (and (not (set-member? var const-proc-vars))
                                         (not (set-member? var clo-vars))))
                                  vars)))
                    (liv (set-union live (free-variables node))))
                (let loop3 ((vars* other-vars-list))
                  (if (not (null? vars*))
                    (let* ((var (first vars*))
                           (val (var->val var))
                           (needed (vals-live-vars liv
                                     (map var->val (after-first vars*)))))
                      (if (var-useless? var)
                        (gen-node val needed 'side)
                        (begin
                          (save-val (gen-node val needed 'need) var needed)
                          (if (mutable? var)
                            (emit-make-cell (var->opnd* var) needed))))
                      (loop3 (after-first vars*)))))
                (gen-node node live why)))))))))

(define (save-arg opnd var live)
  (if (glob? opnd)
    (add-constant-var var opnd)
    (save-val opnd var live)))

(define (save-val opnd var live)
  (cond ((or (imm? opnd) (prim? opnd) (lbl? opnd))
         (add-constant-var var opnd))
        ((and (reg? opnd)
              (not (set-member? (reg-num opnd) (live-regs live))))
         (put-var opnd var))
        ((and (slot? opnd)
              (not (set-member? (slot-num opnd) (live-slots live))))
         (put-var opnd var))
        (else
         (let ((slot (lowest-dead-slot live)))
           (emit-copy opnd slot live)
           (put-var slot var)))))

(define (save-var opnd var live)
  (cond ((or (imm? opnd) (prim? opnd) (lbl? opnd))
         (add-constant-var var opnd)
         var)
        ((or (glob? opnd) (reg? opnd) (slot? opnd))
         (get-var opnd))
        (else
         (let ((dest (or (highest-dead-reg live) (lowest-dead-slot live))))
           (emit-copy opnd dest live)
           (put-var dest var)
           var))))

(define (var-useless? var)
  (and (set-empty? (var-refs var))
       (set-empty? (var-sets var))))

(define (vals-live-vars live vals)
  (if (null? vals)
    live
    (vals-live-vars (set-union live (free-variables (first vals)))
                    (after-first vals))))

;------------------------------------------------------------------------------

(define (emit-label num)
  (put-instr (list 'label num)))

(define (emit-entry-label num)
  (put-instr (list 'entry-label num)))

(define (emit-check-stack)
  (put-instr '(check-stack)))

(define (emit-enter rest? closed? nb-args)
  (put-instr (list 'enter rest? closed? nb-args)))

(define (emit-if not-dir cnd src label live)
  (let ((src-opnd (emit-operands src live #f)))
    (put-instr (list 'if not-dir cnd src-opnd label))))

(define (emit-jump check? dealloc-to nb-args src live)
  (let* ((src-opnd (emit-source src live))
         (dealloc (- frame-size dealloc-to)))
    (dealloc-slots dealloc)
    (put-instr (list 'jump check? dealloc nb-args src-opnd))))

(define (emit-call check? dealloc-to nb-args src live)
  (let* ((src-opnd (emit-source src live))
         (dealloc (- frame-size dealloc-to)))
    (dealloc-slots dealloc)
    (put-instr (list 'call check? dealloc nb-args src-opnd))))

(define (emit-open-code prim args dest live)
  (let ((opnds* (emit-operands (append args (list dest)) live #t)))
    (put-instr (list 'open-code prim opnds*))))

(define (emit-make-cell src live)
  (let ((src-opnd (emit-source src live)))
    (put-instr (list 'make-cell src-opnd))))

(define (emit-make-closures sizes)
  (alloc-slots (length sizes))
  (put-instr (list 'make-closures sizes)))

(define (emit-dealloc n)
  (if (> n 0)
    (begin
      (dealloc-slots n)
      (put-instr (list 'dealloc n)))))

(define (emit-copy src dst live)
  (if (not (equal? src dst))
    (let* ((opnds (emit-operands (list src dst) live #t)))
      (put-instr (list 'copy (first opnds) (second opnds))))))

(define (slot-num-in opnd)
  (cond ((slot? opnd) (slot-num opnd))
        ((indx? opnd) (slot-num-in (indx-base opnd)))
        (else         -1)))

(define (replace-slot-in opnd new)
  (cond ((slot? opnd) new)
        ((indx? opnd) (make-indx (replace-slot-in (indx-base opnd) new)
                                 (indx-offset opnd)))
        (else         (error "operand does not contain a slot reference"))))

(define (emit-source opnd live)
  (first (emit-operands (list opnd) live #f)))

(define (emit-operands opnds live last-is-dest?)
  (let ((highest-live-slot (or (highest (live-slots live)) 0)))
    (define (emit-opnds opnds first?)
      (if (null? opnds)
        '()
        (let* ((opnd (first opnds))
               (new-opnd
                (if (and (slot? opnd) (> (slot-num opnd) frame-size))
                  (begin (push-slot) push-opnd)
                  (let ((opnd-slot (slot-num-in opnd)))
                    (if (< opnd-slot 0)
                      opnd
                      (let ((highest-slot-after
                             (apply max
                                    highest-live-slot
                                    (map slot-num-in (after-first opnds)))))
                        (if (and (>= opnd-slot highest-slot-after) first?)
                          (begin
                            (emit-dealloc (- frame-size opnd-slot))
                            (set! first? #f)))
                        (replace-slot-in opnd
                          (if (and (> opnd-slot highest-slot-after)
                                   (or (not (null? (after-first opnds)))
                                       (not last-is-dest?)))
                            (begin (pop-slot) pop-opnd)
                            (make-stk (+ (- frame-size opnd-slot) 1))))))))))
	  (cons new-opnd (emit-opnds (after-first opnds) first?)))))
  (emit-opnds opnds #t)))

(define (put-instr instr)
;(newline) (display "--- ") (write instr)
  (queue-put! instr-queue instr))

;------------------------------------------------------------------------------
