; file: "compiler.scm"

;------------------------------------------------------------------------------
;
; The file compiler:
; -----------------

(define (compile-file source . optional)

  (let* ((source-port (open-input-file (string-append source ".scm")))
         (target-port (open-output-file (string-append source ".s")))
         (asm-output? (not (null? optional)))
         (asm-port (and asm-output?
                        (if (string? (car optional))
                          (open-output-port (car optional))
                          (current-output-port)))))

  (define (parse expression env)
    (if (define-expr? expression)
      (let ((name (definition-variable expression)))
        (newline) (display "parsing define to ") (write name)
        (list 'define (lookup-var name env)
          (expression->parse-tree (definition-value expression) env)))
      (begin
        (newline) (display "parsing an expression")
        (list 'eval (expression->parse-tree expression env)))))

  (define (read-program port env)
    (let ((expression (read port)))
      (if (eof-object? expression)
        '()
        (let ((x (parse expression env)))
          (cons x (read-program port env))))))

  (define (dump-objects)
    (write-code "|--------------------------------")
    (do ((l (list-tail (queue->list object-queue)
                       (length initial-object-queue))
            (after-first l)))
        ((null? l))
      (write-code)
      (let ((x (first l)))
        (let ((obj (car x)) (num (cdr x)))
          (cond ((pair? obj)
                 (write-code "PAIR_OBJECT(obj" num ","
                             (encode-object (car obj)) ","
                             (encode-object (cdr obj)) ")"))
                ((symbol? obj)
                 (write-code "SYMBOL_OBJECT(obj" num ","
                             (encode-object (symbol->string obj)) ")"))
                ((string? obj)
                 (write-code "STRING_OBJECT(obj" num ")")
                 (for-each
                   (lambda (l) (write-code tab ".byte" tab (list-of l ",")))
                   (partition (map char->integer (string->list obj)) 10))
                 (write-code "END"))
                ((vector? obj)
                 (write-code "VECTOR_OBJECT(obj" num ")")
                 (for-each
                   (lambda (l) (write-code tab ".long" tab (list-of l ",")))
                   (partition (map encode-object (vector->list obj)) 10))
                 (write-code "END"))
                (else
                 (error "unknown object kind"))))))
    (write-code)
    (write-code "PAIR_OBJECT(symbol_table,symbol_table1,"
                (encode-object symbol-table)
                ")")
    (write-code "SYMBOL_OBJECT(symbol_table1,symbol_table2)")
    (write-code "STRING_OBJECT(symbol_table2)")
    (write-code "END"))

  (set! *global-env* (make-standard-environment))
  (let ((infile (string-append filename ".scm"))
        (outfile (string-append filename ".s")))
    (let ((input (open-input-file infile)))
      (set! *code-port* (open-output-file outfile))
      (let ((program (read-program input *global-env*)))
        (set! object-counter (make-counter initial-object-counter))
        (set! object-queue (list->queue (append initial-object-queue '())))
        (set! symbol-table initial-symbol-table)
        (newline)
        (compile-procedures program)
        (compile-top-level program)
        (newline)
        (newline) (display "dumping objects")
        (dump-objects)
        (newline))
      (close-input-port input)
      (close-output-port *code-port*))))

(define (encode-object obj)
  (let ((obj-repr (object-repr obj)))
    (if obj-repr
      obj-repr
      (list "obj" (add-object obj)))))

(define *global-env* '())

(define (make-standard-environment)
  (let ((env (make-empty-environment)))
    (set! symbol-table-var (lookup-var 1 env))
    (for-each (lambda (name) (lookup-var name env))
              (map first *primitive-procedures*))
    env))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Generate code for an expression:
; -------------------------------

; Generate code to evaluate the expression whose parse tree is 'node' for
; the reason specified by 'reason' and without loosing any of the values that
; are still in 'live'.  Values in 'live' which are on the stack will not be
; moved unless 'dest' is such that it would be overwritten.  'reason' can be
; one of the following:
;
;   SIDE     the expression is evaluated for its side effect and the
;            result is not needed
;
;   TAIL     the expression is evaluated in tail position and the result
;            must be passed to the continuation in the result register
;
;   INTO     the expression is evaluated for its value (and side effect)
;            and the result must be stored in the location specified by 'dest'
;
;   KEEP     the expression is evaluated for its value (and side effect)
;            and the result must be kept; if the result must be stored
;            somewhere, then it is stored in the location specified by 'dest'
;            if it isn't used or else in a register or else on the stack
;
; *bb* and *bb-set* are side effected
;
; note: 'dest' can be a VM opnd or a symbolic name (i.e. var node or pt node)

(define (gen-expr! node reason dest live)

  (define (handle-result vm-opnd)
    reason dest)

  (cond ((cst? node)
         (let ((val (cst-val node)))
           (handle-result
             (if (prim-proc? val)
               (make-prim (prim-proc->entry val #f))
               (make-obj val)))))

        ((ref? node)
         (handle-result (name->vm-opnd (ref-var node))))

        ((set? node)
         (gen-expr! (set-val node) 'INTO (set-var node) live)
         (handle-result result-reg))

        ((cond? node)
         (gen-cond (cond-pre node) (cond-con node) (cond-alt node) live why))

        ((proc? node)
         (let* ((closed (not-constant-closed-vars node))
                (closed-list (set->list closed))
                (proc-label (schedule-gen-proc node closed-list)))
           (let ((opnd
                  (if (null? closed-list)
                    (begin
                      (add-known-proc (lbl-num proc-label) node)
                      proc-label)
                    (begin
                      (emit-dealloc (- frame-size
                                       (slot-num (highest-live-slot
                                                   (set-union closed live)))))
                      (emit-make-closures (list (length closed-list)))
                      (let ((slot (make-slot frame-size))
                            (var (make-arg-var 'closure)))
                        (put-var slot var)
                        (prepare-closure var proc-label closed-list live)
                        slot)))))
             (gen-return opnd why))))

        ((call? node)
         (gen-call node live why))

        (else
         (error "Unknown parse tree node type"))))

(define (gen-return opnd why)
  (cond ((eq? why 'tail)
         (emit-copy opnd result-reg cont-var-set)
         (emit-jump #f 0 #f (var->opnd cont-var) set-empty))
        (else
         opnd)))

(define (not-constant-closed-vars proc)
  (set->list (set-keep not-constant-var? (closed-variables proc))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (nb-args-passed-on-stack nb-args)
  (max 0 (- nb-args nb-regs)))

(define (nb-args-passed-in-regs nb-args)
  (min nb-args nb-regs))

(define (operator? arg-num)
  (= arg-num 0))

(define (arg-passed-on-stack? arg-num nb-on-stack)
  (and (> arg-num 0) (<= arg-num nb-on-stack)))

(define (arg-passed-in-reg? arg-num nb-on-stack)
  (> arg-num nb-on-stack))

(define (open-codable-with-simple-args? node)
  (and (call? node)
       (ref? (call-oper node))
))



