;------------------------------------------------------------------------------
;
; Context manipulation:
; --------------------

; A context is a description of the runtime 'environment' of the virtual
; machine at a given point in time.  It is used to find out how a value
; (for example, a variable) can be accessed.  Thus a context is a map
; from 'value tags' to virtual machine operands.  A value tag can be one
; of the following:
;
;   a constant object - the operand is the value of the constant
;   a variable        - the operand is the value of the variable
;   a parse tree node - the operand is the value of the expression
;                       associated with this node
;
; Contexts track the location of values as they are moved around and also keep
; track of the copies that are made of values.

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Values can be copied from one location to another and thus a value tag
; can be accessed more than one way.  The procedure 'val-tag->vm-opnds'
; returns a list of all the ways the value tag passed as an argument can be
; accessed in the current context.

(define (val-tag->vm-opnds val-tag)
  (cond ((obj? val-tag)
         (let ((x (assq val-tag *context-csts*)))
           (if x (cdr x)
             (let ((y (list val-tag)))
               (set! *context-csts* (cons (cons val-tag y) *context-csts*))
               y))))
        ((var? val-tag)
         (if (global? val-tag)
           (list (make-glob 123))
           (cdr (assq val-tag *context-vars*))))
        (else
         (cdr (assq val-tag *context-nodes*)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; (val-tag-can-be-found-in val-tag vm-opnds) is used to initially associate
; a value tag to a set of virtual machine operands in the current context.

(define (val-tag-can-be-found-in val-tag vm-opnds)
  (cond ((var? val-tag)
         (if (not (global? val-tag))
           (set! *context-vars*
             (cons (cons val-tag vm-opnds) *context-vars*))))
        ((cst? val-tag)
         (let ((val (cst-val val-tag)))
           (set! *context-csts*
             (cons (cons val vm-opnds) *context-csts*))))
        ((ref? val-tag)
         (val-tag-can-be-found-in (ref-var val-tag) vm-opnds))
        (else
         (set! *context-nodes*
           (cons (cons val-tag vm-opnds) *context-nodes*)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; (vm-opnd-is-now-undefined vm-opnd) is used to indicate that a virtual
; machine operand no longer contains the value it had.

(define (vm-opnd-is-now-undefined vm-opnd)

  (define (remove x l)
    (cond ((null? l)       '())
          ((eq? x (car l)) (cdr l))
          (else            (cons (car l) (remove x (cdr l))))))

  (define (undefine-in l)
    (if (not (null? l))
      (let ((x (car l)))
        (if (memq vm-opnd (cdr x)) (set-cdr! x (remove vm-opnd (cdr x))))
        (undefine-in (cdr l)))))

  (undefine-in *context-vars*)
  (undefine-in *context-csts*)
  (undefine-in *context-nodes*))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; (vm-opnd-has-been-copied-in vm-opnd vm-loc) is used to indicate that a copy
; of a virtual machine operand has been put in another location.

(define (vm-opnd-has-been-copied-in vm-opnd vm-loc)

  (define (add-copy-in l)
    (if (not (null? l))
      (let ((x (car l)))
        (if (memq vm-opnd (cdr x)) (set-cdr! x (cons vm-loc (cdr x))))
        (add-copy-in (cdr l)))))

  (if (not (or (mem? vm-loc) (glob? vm-loc)))
    (begin
      (vm-loc-is-now-undefined vm-loc)
      (add-copy-in *context-vars*)
      (add-copy-in *context-csts*)
      (add-copy-in *context-nodes*))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; (copy-val-tag-to val-tag dest) generates the code to copy the value tag
; to a virtual machine location.

(define (copy-val-tag-to val-tag dest)
  (let ((srcs (val-tag->vm-opnds val-tag)))
    (if (not (memq dest srcs))
      (begin
        (emit-COPY ...)
(vm-opnd-is-now-undefined dest)
(val-tag-can-now-be-found-in-vm-opnd val-tag dest)
        
COPY(src-val-tag,dst)

dst -> dst*



; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; (gen-expr! node reason dest live)
;
; TAIL
; SIDE
; INTO
; HINT

(define (gen-expr! node reason dest live)

  (cond ((cst? node)
         (case reason
           ((TAIL))
           ((SIDE))
           ((INTO))
           ((HINT))))

        ((ref? node)
         (gen-return
           (cond ((eq? why 'side)
                  result-reg)
                 ((and assume-standard-prims
                       (global? (ref-var node))
                       (primitive-procedure (var-name (ref-var node))))
                  (make-prim (var-name (ref-var node)) #f))
                 (else
                  (var->opnd (ref-var node))))
           why))

        ((set? node)
         (let* ((src (gen-node (set-val node)
                               (set-adjoin live (set-var node))
                               'keep))
                (dst (var->opnd (set-var node))))
           (emit-copy src dst live))
         (gen-return result-reg why))

        ((cond? node)
         (gen-cond (cond-pre node) (cond-con node) (cond-alt node) live why))

        ((proc? node)
         (let* ((closed (not-constant-closed-vars node))
                (closed-list (set->list closed))
                (proc-label (schedule-gen-proc node closed-list)))
           (let ((opnd
                  (if (null? closed-list)
                    (begin
                      (add-known-proc (lbl-num proc-label) node)
                      proc-label)
                    (begin
                      (emit-dealloc (- frame-size
                                       (slot-num (highest-live-slot
                                                   (set-union closed live)))))
                      (emit-make-closures (list (length closed-list)))
                      (let ((slot (make-slot frame-size))
                            (var (make-arg-var 'closure)))
                        (put-var slot var)
                        (prepare-closure var proc-label closed-list live)
                        slot)))))
             (gen-return opnd why))))

        ((call? node)
         (gen-call node live why))

        (else
         (error "Unknown parse tree node type"))))

