; file: "virtual.scm"

;------------------------------------------------------------------------------
;
; Virtual machine abstraction:

; The virtual machine model is basically a set of instructions and a set of
; operand types.  The operand types are divided into 2 classes: locations and
; values.  Locations can be used both as source and destination operands in
; instructions whereas values can only be used as source operands.  A location
; can be either one of: a register, a stack slot, a global variable or a memory
; location.  Memory locations correspond to heap allocated objects and are only
; accessible through the dereferencing of an object pointer.
;
;  registers                       stack             global variables
;  ________                                              ________
; |________| 0 (continuation)    |________| .  top      |________| 1
; |________| 1                   |________| .           |________| 2
; |________| 2                   |________| .  /|\      |________| .
; |________| ...                 |________| 2   |       |________| .
; |________| nb-regs             |________| 1   |       |________| .

; for the definition of 'nb-regs' see file "target.scm"

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Virtual machine operands:
; ------------------------

; Operands are represented with small integers and cons cells.  It is assumed
; that when an operand is numerically encoded it fits in a fixnum.  This makes
; it possible to implement quick operations on sets of operands because
; operands can be tested for equality using 'eq?'.  The encoding is as follows:
;
; encoding               operand type
;
; 10001 ..               global variable
; 0 .. 10000             register
; -1 .. -9999            stack slot
; -10000                 push to / pop from stack
; -10001 .. -19999       local label
; -20000 ..              primitive procedure
; (base . offset)        memory location
; (object)               scheme object

(define (identical-vm-opnds? opnd1 opnd2) (eq? opnd1 opnd2))

; Locations:
; ---------

; -- location is a register (reg 1 is 1st general reg)
(define (make-reg num) num)
(define (reg? x) (and (integer? x) (>= x 0) (<= x 10000)))
(define (reg-num x) x)

; -- location is in the stack (first slot in procedure's frame is number 1)
(define (make-stk num) (- num))
(define (stk? x) (and (integer? x) (> x -10000) (< x 0)))
(define (stk-num x) (- x))

; -- location is to be allocated or deallocated from the top of the stack
; (depending on whether it is used as a source or destination location)
(define (make-push/pop) -10000)
(define (push/pop? x) (eq? x -10000))

; -- location is a global variable
(define (make-glob num) (+ num 10000))
(define (glob? x) (and (integer? x) (> x 10000)))
(define (glob-num x) (- x 10000))

; -- location is a memory location at a certain byte offset from a base
(define (make-mem base offset) (encode-mem base offset))
(define (mem? x) (and (pair? x) (integer? (cdr x))))
(define (mem-base x) (car x))
(define (mem-offset x) (cdr x))

; Values:
; ------

; -- value is the address of a local label
(define (make-lbl num) (- -10000 num))
(define (lbl? x) (and (integer? x) (> x -20000) (< x -10000)))
(define (lbl-num x) (- -10000 x))
(define label-limit 10000)

; -- value is the address of a primitive procedure's entry point
(define (make-prim entry) (- -20000 entry))
(define (prim? x) (and (integer? x) (<= x -20000)))
(define (prim-entry x) (- -20000 x))

; -- value is a scheme object
(define (make-obj val) (encode-obj val))
(define (obj? x) (and (pair? x) (null? (cdr x))))
(define (obj-val x) (car x))

; Common operands:
; ---------------

(define continuation-reg (make-reg 0)) ; the continuation register
(define result-reg (make-reg 1)) ; the result register
(define push/pop (make-push/pop))

; Operand writing:
; ---------------

(define (write-vm-opnd vm-opnd source? port)
  (cond ((reg? vm-opnd)
         (display "reg(" port)
         (display (reg-num vm-opnd) port)
         (display ")" port))
        ((stk? vm-opnd)
         (display "stk(" port)
         (display (stk-num vm-opnd) port)
         (display ")" port))
        ((push/pop? vm-opnd)
         (display (if source? "pop" "push") port))
        ((glob? vm-opnd)
         (display "glob(" port)
         (display (glob-num vm-opnd) port)
         (display ")" port))
        ((mem? vm-opnd)
         (display "mem(" port)
         (write-vm-opnd (mem-base vm-opnd) #t port)
         (display "," port)
         (display (mem-offset vm-opnd) port)
         (display ")" port))
        ((lbl? vm-opnd)
         (display "lbl(" port)
         (display (lbl-num vm-opnd) port)
         (display ")" port))
        ((prim? vm-opnd)
         (display "prim(" port)
         (display (prim-entry vm-opnd) port)
         (display ")" port))
        ((obj? vm-opnd)
         (display "obj(" port)
         (write (obj-val vm-opnd) port)
         (display ")" port))
        (else
         (compiler-internal-error
           "write-vm-opnd, unknown 'vm-opnd'"
           vm-opnd))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Virtual machine instructions:
; ----------------------------

; The virtual machine's instruction set is composed of the following
; instructions:

; LABEL( lbl_num [,nb_parms [,REST] [,CLOSED]] )
;
; -- declare the local label `lbl_num'; if `nb_parms' is present, the label is
; -- `first class' in that it can be used as any other Scheme value (in fact
; -- it is of procedure type and `nb_parms' is the number of parameters
; -- expected by the procedure); if `REST' is present, the last parameter
; -- is a `rest' parameter; if `CLOSED' is present, the procedure is
; -- actually the body of a true closure
; --
; -- note: first class labels must be jumped to by a JUMP instruction which
; -- specifies the number of arguments

; CHECK_STACK
;
; -- grow stack if it is almost full and poll interrupts

; DEALLOC( n )
;
; -- remove `n' values from stack

; IF( test, opnd1, [opnd2,] lbl_num ) / IF_NOT( test, opnd1, [opnd2,] lbl_num )
;
; -- jump to label `lbl_num' if the `test' is true (or not true) of the
; -- operand(s) (`test' can be: TRUE?, NULL?, FIXNUM?, FLONUM?, CHAR?, PAIR?,
; -- HUNK?, PROCEDURE?, STRING?, VECTOR?, SYMBOL?, EQ?, =, CHAR=?, <, CHAR<?,
; -- >, CHAR>?, <=, CHAR<=?, >=, CHAR>=?)

; JUMP( opnd, n [,nb_args [,CHECK]] )
;
; -- jump to the address `opnd' and deallocate `n' values from the stack
; -- (if `nb_args' is present, pass it as an argument count; if `CHECK' is
; -- present, `opnd' is first checked to make sure it is a procedure)

; OPEN_CODE( prim, opnd1, ..., opndn [,loc] )
;
; -- compute the primitive `prim' applied to the operands and, if `loc' is
; -- specified, put the result in the location `loc'
; -- (for example: OPEN_CODE( prim(+,2), reg(1), reg(2), reg(1) ) )

; COPY( opnd, loc )
;
; -- copy a value from `opnd' to `loc'; COPY is equivalent to
; -- OPEN_CODE( prim(IDENTITY,1), opnd, loc )

; MAKE_CLOSURES( size1:loc1, ..., sizen:locn )
;
; -- create `n' closure objects (each having the corresponding number of slots
; -- for closed variables) and put their addresses in the corresponding
; -- locations

; note: when multiple operands appear in an instruction, it is assumed that
; they are accessed from left to right.

; Virtual instructions have a linear structure.  However, this is not how
; they are put together to form a piece of code.  Rather, virtual instructions
; are grouped into 'basic blocks' which are 'linked' together.  A basic block
; is a LABEL instruction followed by a sequence of non-branching instructions
; (i.e. CHECK_STACK, DEALLOC, OPEN_CODE, COPY or MAKE_CLOSURES) terminated by
; a single branch instruction (i.e. IF, IF_NOT or JUMP).  Links between basic
; blocks are denoted using label references.  When a basic block ends with an
; IF or IF_NOT instruction, the block will be linked to the two basic blocks
; corresponding to the two possible control paths out of the IF or IF_NOT
; instruction.  When a basic block ends with a JUMP instruction, there is
; either zero or one link.
;
; Basic blocks naturally group together to form 'basic block sets'.  A basic
; block set describes the code of a procedure.

(define (make-counter limit limit-error)
  (let ((count 0))
    (lambda ()
      (if (< count limit)
        (begin (set! count (+ count 1)) count)
        (limit-error)))))

(define (make-basic-block-set)
  (vector (make-counter label-limit ; 0 - label counter
            (lambda ()
              (compiler-limitation-error
                "procedure is too long [too many labels]")))
          (queue-empty)))           ; 1 - basic block queue

(define (basic-block-set-lbl-counter bb-set)
  (vector-ref bb-set 0))

(define (new-basic-block-lbl! bb-set)
  ((basic-block-set-lbl-counter bb-set)))

(define (make-basic-block lbl nb-parms rest? closed? frame-size bb-set)
  (let ((bb (vector
              frame-size    ; 0 - frame size after LABEL instruction
              (list 'LABEL lbl nb-parms rest? closed?) ; 1 - LABEL instruction
              (queue-empty) ; 2 - sequence of non-branching instrs
              '()           ; 3 - branch instruction
              '()           ; 4 - fall through label
              '())))        ; 5 - labels of the precedents
    (queue-put! (vector-ref bb-set 1) bb)
    bb))

(define (basic-block-frame-size bb)   (vector-ref bb 0))
(define (basic-block-label bb)        (vector-ref bb 1))
(define (basic-block-lbl bb)          (cadr (vector-ref bb 1)))
(define (basic-block-non-branch bb)   (queue->list (vector-ref bb 2)))
(define (basic-block-branch bb)       (vector-ref bb 3))
(define (basic-block-cont-lbl bb)     (vector-ref bb 4))
(define (basic-block-precedents bb)   (vector-ref bb 5))

(define (append-CHECK_STACK! basic-block)
  (queue-put! (vector-ref basic-block 2) '(CHECK_STACK)))

(define (append-DEALLOC! basic-block n)
  (queue-put! (vector-ref basic-block 2) (list 'DEALLOC n)))

(define (append-OPEN_CODE! basic-block prim opnds loc)
  (queue-put! (vector-ref basic-block 2) (list 'OPEN_CODE prim opnds loc)))

(define (append-COPY! basic-block opnd loc)
  (queue-put! (vector-ref basic-block 2) (list 'COPY opnd loc)))

(define (append-MAKE_CLOSURES! basic-block sizes-locs)
  (queue-put! (vector-ref basic-block 2) (list 'MAKE_CLOSURES sizes-locs)))

(define (append-IF! basic-block test opnd1 opnd2 true-lbl false-lbl)
  (vector-set! basic-block 3 (list 'IF test opnd1 opnd2 true-lbl))
  (vector-set! basic-block 4 false-lbl))

(define (append-IF_NOT! basic-block test opnd1 opnd2 true-lbl false-lbl)
  (vector-set! basic-block 3 (list 'IF_NOT test opnd1 opnd2 true-lbl))
  (vector-set! basic-block 4 false-lbl))

(define (append-JUMP! basic-block opnd n nb-args check?)
  (vector-set! basic-block 3 (list 'JUMP opnd n nb-args check?))
  (vector-set! basic-block 4 #f))

(define (add-precedent-block! basic-block lbl)
  (vector-set! basic-block 5 (cons lbl (vector-ref basic-block 5))))

; Linearization of basic block sets:
; ---------------------------------

(define (basic-block-set->list bb-set entry-lbl)
  (let ((bb-list (queue->list (vector-ref bb-set 1)))
        (bb-queue (queue-empty)))

    (define (find-block lbl)
      (let loop ((l bb-list))
        (if (= (basic-block-lbl (car l)) lbl)
          (car l)
          (loop (cdr l)))))

    (define (remove-bb bb bb-left)
      (if (eq? (car bb-left) bb)
        (cdr bb-left)
        (cons (car bb-left) (remove-bb bb (cdr bb-left)))))

    ; return a basic block which ends with a branch to 'bb' (and that is
    ; still in 'bb-left') or #f if there aren't any
    (define (prec-bb bb bb-left)
      (let loop ((l (basic-block-precedents bb)))
        (if (null? l)
          #f
          (let ((prec (find-block (car l))))
            (if (memq prec bb-left) prec (loop (cdr l)))))))

    ; return the basic block which 'bb' jumps to (and that is still in
    ; 'bb-left') or #f if there aren't any
    (define (next-bb bb bb-left)
      (let* ((branch (basic-block-branch bb))
             (branch-name (car branch))
             (next (if (eq? branch-name 'JUMP)
                     (let ((dest (cadr branch)))
                       (and (lbl? dest) (find-block (lbl-num dest))))
                     (let ((cont-block (find-block (basic-block-cont-lbl bb))))
                       (if (memq cont-block bb-left)
                         cont-block
                         (find-block (car (cddddr branch))))))))
        (and next (memq next bb-left) next)))

    (define (insert-now! bb bb-left)
      (queue-put! bb-queue bb)
      (insert-after! bb (remove-bb bb bb-left)))

    (define (insert-after! bb bb-left)
      (if (not (null? bb-left))
        (if bb
          (let ((next (next-bb bb bb-left)))
            (insert-later! next bb-left insert-after!))
          (insert-now! (car bb-left) bb-left))))

    (define (insert-later! bb bb-left do!)
      (if bb
        (let ((bb-left (remove-bb bb bb-left)))
          (insert-later! (prec-bb bb bb-left) bb-left
            (lambda (x bb-left) (queue-put! bb-queue bb) (do! bb bb-left))))
        (do! #f bb-left)))

    (insert-now! (find-block entry-lbl) bb-list)

    (queue->list bb-queue)))

; Target code generation of a basic block list:
; --------------------------------------------

; this procedure uses the following procedures of the 'target' package:
;   target-basic-block-begin!
;   target-generate-vm-instr!
;   target-basic-block-end!

(define (generate-basic-block-list! basic-block-list)

  (define (generate-basic-block! prec-lbl bb next-lbl)

    (define (general-label? label) (caddr label)) ; nb_parms specified?
    (define (general-jump? jump) (cadddr jump)) ; nb_args specified?

    (target-basic-block-begin! bb) ; tell target generator we are starting a bb

    (let ((label (basic-block-label bb))
          (precedents (basic-block-precedents bb)))
      (if (not (and (not (general-label? label))
                    (= (length precedents) 1)
                    (eq? (car precedents) prec-lbl)))
        (target-generate-vm-instr! label)))

    (for-each target-generate-vm-instr! (basic-block-non-branch bb))

    (let* ((branch (basic-block-branch bb))
           (branch-name (car branch)))
      (if (eq? branch-name 'JUMP)
        (let ((dest (cadr branch)))
          (if (and (not (general-jump? branch))
                   (lbl? dest)
                   (eq? (lbl-num dest) next-lbl))
            (if (> (caddr branch) 0)
              (target-generate-vm-instr! (list 'DEALLOC (caddr branch))))
            (target-generate-vm-instr! branch)))
        (if (eq? (car (cddddr branch)) next-lbl)
          (target-generate-vm-instr!
            (list (if (eq? branch-name 'IF) 'IF_NOT 'IF)
                  (cadr branch)
                  (caddr branch)
                  (cadddr branch)
                  (basic-block-cont-lbl bb)))
          (begin
            (target-generate-vm-instr! branch)
            (if (not (eq? (basic-block-cont-lbl bb) next-lbl))
              (target-generate-vm-instr!
                (list 'JUMP (make-lbl (basic-block-cont-lbl bb)) 0 #f #f)))))))

    (target-basic-block-end! bb)) ; tell target generator we have ended a bb

  (let loop ((l basic-block-list) (prev #f))
    (if (not (null? l))
      (let ((pres (car l)))
        (generate-basic-block! prev pres
          (if (null? (cdr l)) #f (basic-block-lbl (cadr l))))
        (loop (cdr l) (basic-block-lbl pres))))))

; Instruction writing:
; -------------------

(define (write-vm-instr vm-instr port)
  (let ((name (car vm-instr)) (parms (cdr vm-instr)))
    (cond

      ((eq? name 'LABEL)
       (let ((lbl (car parms))
             (nb-parms (cadr parms))
             (rest? (caddr parms))
             (closed? (cadddr parms)))
         (display "LABEL(" port) (display lbl port)
         (if nb-parms
            (begin
              (display "," port) (display nb-parms port)
              (if rest? (display ",REST" port))
              (if closed? (display ",CLOSED" port))))
         (display ")" port)
         (newline port)))

      ((eq? name 'CHECK_STACK)
       (display "CHECK_STACK" port)
       (newline port))

      ((eq? name 'DEALLOC)
       (let ((n (car parms)))
         (display "DEALLOC(" port) (display n port) (display ")" port)
         (newline port)))

      ((eq? name 'OPEN_CODE)
       (let ((prim (car parms))
             (opnds (cadr parms))
             (loc (caddr parms)))
         (display "OPEN_CODE(" port)
         (write-vm-opnd prim #t port)
         (for-each (lambda (x) (display "," port) (write-vm-opnd x #t port))
                   opnds)
         (if loc (begin (display "," port) (write-vm-opnd loc #f port)))
         (display ")" port)
         (newline port)))

      ((eq? name 'COPY)
       (let ((vm-opnd (car parms))
             (vm-loc (cadr parms)))
         (display "COPY(" port)
         (write-vm-opnd vm-opnd #t port)
         (display "," port)
         (write-vm-opnd vm-loc #f port)
         (display ")" port)
         (newline port)))

      ((eq? name 'MAKE_CLOSURES)
       (let ((sizes-locs (car parms)))
         (define (write-size-loc size-loc port)
           (display (car size-loc) port)
           (display ":" port)
           (write-vm-opnd (cdr size-loc) #f port))
         (display "MAKE_CLOSURES(" port)
         (write-size-loc (car sizes-locs) port)
         (for-each (lambda (x) (display "," port) (write-size-loc x port))
                   (cdr sizes-locs))
         (display ")" port)
         (newline port)))

      ((or (eq? name 'IF) (eq? name 'IF_NOT))
       (let ((test (car parms))
             (vm-opnd1 (cadr parms))
             (vm-opnd2 (caddr parms))
             (lbl (cadddr parms)))
         (display (if (eq? name 'IF) "IF(" "IF_NOT(") port)
         (display test port)
         (display "," port) (write-vm-opnd vm-opnd1 #t port)
         (if vm-opnd2
           (begin (display "," port) (write-vm-opnd vm-opnd2 #t port)))
         (display "," port) (display lbl port)
         (display ")" port)
         (newline port)))

      ((eq? name 'JUMP)
       (let ((vm-opnd (car parms))
             (n (cadr parms))
             (nb-args (caddr parms))
             (check? (cadddr parms)))
         (display "JUMP(" port)
         (write-vm-opnd vm-opnd #t port)
         (display "," port) (display n port)
         (if nb-args
           (begin
             (display "," port) (display nb-args port)
             (if check? (display ",CHECK" port))))
         (display ")" port)
         (newline port)))

      (else
       (compiler-internal-error "write-vm-instr, unknown 'name'" name)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (virtual-begin!) ; initialize package
  (set! *mem-table* '())
  (set! *mem-count* 0)
  (set! *obj-table* '())
  (set! *obj-count* 0))

(define (virtual-end!) ; finalize package
  (set! *mem-table* '())
  (set! *obj-table* '()))

;------------------------------------------------------------------------------
;
; Stuff local to the package:

(define *mem-table* '()) ; keeps all previously generated 'mem' locations
(define *obj-table* '()) ; keeps all previously generated 'obj' values

(define (encode-mem base offset)
  (let loop ((l *mem-table*))
    (if (null? l)
      (begin
        (set! *mem-table* (cons (cons base offset) *mem-table*))
        (car *mem-table*))
      (let ((x (car l)))
        (if (and (identical-vm-opnds? base (car x)) (= offset (cdr x)))
          x
          (loop (cdr l)))))))

(define (encode-obj val)
  (let loop ((l *obj-table*))
    (if (null? l)
      (begin
        (set! *obj-table* (cons (list val) *obj-table*))
        (car *obj-table*))
      (let ((x (car l)))
        (if (identical-objects? val (car x)) x (loop (cdr l)))))))

;------------------------------------------------------------------------------

(define (test-virtual)
  (virtual-begin!)
  (let ((bb-set (make-basic-block-set)))
    (let* ((bb1-lbl (new-basic-block-lbl! bb-set))
           (bb2-lbl (new-basic-block-lbl! bb-set))
           (bb1 (make-basic-block bb1-lbl 6 #f #f 2 bb-set)))

(append-CHECK_STACK! bb1)
(append-DEALLOC! bb1 2)
(append-OPEN_CODE! bb1 (make-prim 14)
  (list (make-reg 1) (make-reg 2)) #f)
(append-OPEN_CODE! bb1 (make-prim 16)
  (list (make-reg 1) (make-reg 2)) (make-reg 1))
(append-COPY! bb1 result-reg (make-stk 1))
(append-COPY! bb1 result-reg (make-stk 2))
(append-COPY! bb1 result-reg push/pop)
(append-COPY! bb1 result-reg (make-glob 10))
(append-COPY! bb1 push/pop (make-mem (make-reg 0) 20))
(append-COPY! bb1 result-reg (make-mem (make-reg 1) 20))
(append-COPY! bb1 result-reg (make-mem (make-mem (make-reg 1) 30) 40))
(append-COPY! bb1 (make-lbl bb1-lbl) result-reg)
(append-COPY! bb1 (make-prim 60) push/pop)
(append-COPY! bb1 (make-obj '#(1 2 3)) result-reg)
(append-MAKE_CLOSURES! bb1
  (list (cons 3 (make-reg 1)) (cons 5 (make-reg 2))))
(append-IF_NOT! bb1 'FIXNUM? result-reg #f bb2-lbl bb1-lbl)

      (let ((bb2 (make-basic-block bb2-lbl #f #f #f 3 bb-set)))

(append-JUMP! bb2 (make-lbl bb1-lbl) 1 #f #f)

(add-precedent-block! bb1 bb1-lbl)
(add-precedent-block! bb1 bb2-lbl)
(add-precedent-block! bb2 bb1-lbl)

(target-begin!)

(target-generate-procedure bb-set bb1-lbl)

(target-end!)

))))

