; file: "target.scm"

;------------------------------------------------------------------------------
;
; Target machine abstraction (for MC68020):

; The virtual machine implementation is a mapping of VM instructions
; and operands to MC68020 instructions and operands.  The mapping of
; operands is fairly simple because MC68020 operands form a superset of
; VM operands.  VM registers are mapped to MC68020 registers, the VM stack
; is implemented with the MC68020's stack and global variables are
; implemented by an array of objects.
;
; The MC68020's registers are dedicated as follows:
;
; D0      argument count register (used by JUMP and LABEL instructions)
; D1..D4  VM register 1 to 4 (general registers)
; D5      always = fixnum mask (to implement the IF(FIXNUM?,...) instruction)
; D6      always = pair mask (to implement the IF(PAIR?,...) instruction)
; D7      always = hunk mask (to implement the IF(HUNK?,...) instruction)
; A0      VM register 0 (continuation register)
; A1..A3  temporary registers (to implement VM instructions)
; A4      always = empty list (for 'falsity' tests)
; A5      always = pointer to the runtime system (primitive procedures, etc.)
; A6      always = pointer to the global variable array
; A7      stack pointer
;
; Thus VM operands get encoded into MC68020 operands as follows:
;
; reg(N)    --> A0 (if N = 0) or DN (if N > 0)
; stk(N)    --> A7@(4*(frame_size-N))
; push/pop  --> A7@- or A7@+
; glob(N)   --> A6@(6*N+256)  (1)
; mem(B,O)  --> Ax@(O) (preceded by instruction 'movl encoded(B),Ax')  (2)
; lbl(N)    --> address of LABEL(N)
; prim(E)   --> address of A5@(E)
; obj(V)    --> #xxx (if V is not memory allocated) or PC@(nnn)  (3)
;
; notes: (1) each global variable takes 6 bytes of storage and the array is
;            preceded by a 260 byte long area
;        (2) this is the general form (where 'x' is 1, 2 or 3); there
;            are some optimizations possible (e.g. when B is (reg 0))
;        (3) if the object is not memory allocated, its bit representation
;            is constant and thus it can be used as an immediate operand;
;            if it is memory allocated, the object is put in a table at
;            the end of the procedure and is referenced with pc-relative
;            addressing
;
; The mapping of instructions is more difficult.  First of all, VM instructions
; are of a rather 'high' level compared to MC68020 instructions and it may
; take a sequence of MC68020 instructions to encode a single VM instruction.
; Frequently, special cases can be exploited to shorten the code needed to
; implement a given VM instruction.  Secondly, VM instructions are homogenous
; in the sense that they take operands of any kind; this is hardly the case
; of the MC68020 which, for example, can not compare two stack locations
; in a single 'cmpl' instruction.

; number of registers in the virtual machine
(define nb-regs 4)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Object representation:
; ---------------------

(define pointer-size 4)             ; size of an object pointer
(define closure-proc-offset 2)      ; offset to procedure pointer in a closure
(define closure-first-var-offset 6) ; offset to first closed var in a closure
(define cell-value-offset 2)        ; offset to value in a cell object
(define pair-car-offset 0)          ; offset to the car of a pair
(define pair-cdr-offset -4)         ; offset to the cdr of a pair
(define vector-element0-offset 2)   ; offset to first element of a vector

(define (index-offset element0-offset i) ; return offset of array element 'i'
  (+ element0-offset (* pointer-size i)))

(define (memory-allocated? x)
  (or (pair? x) (vector? x) (symbol? x) (string? x)))

(define (bit-representation x)
  (cond ((fixnum? x)                (* 8 x))
        ((flonum? x)                1) ; not implemented yet...
        ((null? x)                  -3)
        ((identical-objects? x #f)  -1)
        ((identical-objects? x #t)  -5)
        ((char? x)                  (- (* 2 (char->integer x)) 131071))
        (else
         (compiler-internal-error "bit-representation, unknown 'x'" x))))

(define (fixnum? x)
  (and (integer? x) (>= x -268435456) (<= x 268435455)))

(define (flonum? x)
  (real? x))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; MC68020 operands:
; ----------------

; All operands are represented with integers and can be tested for
; equality using 'eq?'.  The representation is similar to the actual bit
; sequence used by the hardware.  This makes for an efficient
; 'opnd->mode/reg' procedure.

(define (identical-opnds? opnd1 opnd2) (eq? opnd1 opnd2))

; -- data register
(define (make-dreg num) num)
(define (dreg? x) (and (>= x 0) (< x 8)))
(define (dreg-num x) x)

; -- address register
(define (make-areg num) (+ num 8))
(define (areg? x) (and (>= x 8) (< x 16)))
(define (areg-num x) (- x 8))

; -- address register indirect
(define (make-ind areg) (+ areg 8))
(define (ind? x) (and (>= x 16) (< x 24)))
(define (ind-areg x) (- x 8))

; -- address register indirect with postincrement
(define (make-pinc areg) (+ areg 16))
(define (pinc? x) (and (>= x 24) (< x 32)))
(define (pinc-areg x) (- x 16))

; -- address register indirect with predecrement
(define (make-pdec areg) (+ areg 24))
(define (pdec? x) (and (>= x 32) (< x 40)))
(define (pdec-areg x) (- x 24))

; -- address register indirect with displacement
(define (make-disp areg offset) (+ (+ areg 32) (* (modulo offset 65536) 8)))
(define (disp? x) (and (>= x 40) (< x 524328)))
(define (disp-areg x) (+ (remainder x 8) 8))
(define (disp-offset x) (quotient (- x 40) 8))

(define (make-disp* areg offset) ; smarter version of 'make-disp'
  (if (= offset 0) (make-ind areg) (make-disp areg offset)))
(define (disp*? x) (or (ind? x) (disp? x)))
(define (disp*-areg x) (if (ind? x) (ind-areg x) (disp-areg x)))
(define (disp*-offset x) (if (ind? x) 0 (disp-offset x)))

; -- pc relative
(define (make-rel lbl offset) (+ 524328 (+ (modulo offset 65536) (* lbl 65536))))
(define (rel? x) (>= x 524328))
(define (rel-lbl x) (quotient (- x 524328) 65536))
(define (rel-offset x) (remainder (- x 524328) 65536))

; -- immediate
(define (make-imm val) (if (< val 0) (* val 2) (- -1 (* val 2))))
(define (imm? x) (< x 0))
(define (imm-val x) (if (even? x) (quotient x 2) (- (quotient x 2))))

; Common operands:
; ---------------

(define arg-count-reg (make-dreg 0)) ; argument count register
(define fixnum-reg    (make-dreg 5)) ; fixnum mask register
(define pair-reg      (make-dreg 6)) ; pair mask register
(define hunk-reg      (make-dreg 7)) ; hunk mask register

(define cont-reg       (make-areg 0)) ; continuation register
(define empty-list-reg (make-areg 4)) ; register containing the empty list
(define rt-reg         (make-areg 5)) ; runtime pointer register
(define gv-reg         (make-areg 6)) ; global variable pointer register
(define sp-reg         (make-areg 7)) ; stack pointer register

(define push-opnd (make-pdec sp-reg))
(define pop-opnd  (make-pinc sp-reg))

(define stk-limit (make-ind gv-reg))

; Operand conversion procedures:
; -----------------------------

(define (opnd->mode/reg opnd)
  (cond ((disp? opnd) (+ 32 (disp-areg opnd))) ; 101 rrr
        ((rel? opnd)  58)                      ; 111 010
        ((imm? opnd)  60)                      ; 111 100
        (else         opnd)))

(define (opnd->reg/mode opnd)
  (let ((x (opnd->mode/reg opnd)))
    (+ (* 8 (remainder x 8)) (quotient x 8))))

(define (opnd->string opnd)
  (cond ((dreg? opnd)
         (string-append "d" (number->string (dreg-num opnd))))
        ((areg? opnd)
         (string-append "a" (number->string (areg-num opnd))))
        ((ind? opnd)
         (string-append (opnd->string (ind-areg opnd)) "@"))
        ((pinc? opnd)
         (string-append (opnd->string (pinc-areg opnd)) "@+"))
        ((pdec? opnd)
         (string-append (opnd->string (pdec-areg opnd)) "@-"))
        ((disp? opnd)
         (string-append (opnd->string (disp-areg opnd))
                        "@("
                        (number->string (disp-offset opnd))
                        ")"))
        ((rel? opnd)
         (string-append "l"
                        (number->string (rel-lbl opnd))
                        "+"
                        (number->string (rel-offset opnd))))
        ((imm? opnd)
         (string-append "#" (number->string (imm-val opnd))))
        (else
         (compiler-internal-error "opnd->string, unknown 'opnd'" opnd))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; MC68020 instructions:
; --------------------

; Opcodes:
; -------

(define <movl>   8192)
(define <asll>  57728)
(define <asrl>  57472)
(define <rts>   20085)
(define <swap>  18496)
(define <bcc>   25600)
(define <bcs>   25856)
(define <beq>   26368)
(define <bge>   27648)
(define <bgt>   28160)
(define <bhi>   25088)
(define <ble>   28416)
(define <bls>   25344)
(define <blt>   27904)
(define <bmi>   27392)
(define <bne>   26112)
(define <bpl>   27136)
(define <bra>   24576)
(define <bvc>   26624)
(define <bvs>   26880)
(define <bsr>   24832)
(define <jmp>   20160)
(define <jsr>   20096)
(define <pea>   18496)
(define <addl>  54972)
(define <addw>  54908)
(define <cmpl>  48316)
(define <subl>  38076)
(define <moveq> 28672)
(define <lea>   16832)
(define <addql> 20608)
(define <subql> 20864)

(define (opcode->string opcode)
  (cond ((= opcode <movl> ) "movl")
        ((= opcode <asll> ) "asll")
        ((= opcode <asrl> ) "asrl")
        ((= opcode <rts>  ) "rts")
        ((= opcode <swap> ) "swap")
        ((= opcode <bcc>  ) "bcc")
        ((= opcode <bcs>  ) "bcs")
        ((= opcode <beq>  ) "beq")
        ((= opcode <bge>  ) "bge")
        ((= opcode <bgt>  ) "bgt")
        ((= opcode <bhi>  ) "bhi")
        ((= opcode <ble>  ) "ble")
        ((= opcode <bls>  ) "bls")
        ((= opcode <blt>  ) "blt")
        ((= opcode <bmi>  ) "bmi")
        ((= opcode <bne>  ) "bne")
        ((= opcode <bpl>  ) "bpl")
        ((= opcode <bra>  ) "bra")
        ((= opcode <bvc>  ) "bvc")
        ((= opcode <bvs>  ) "bvs")
        ((= opcode <bsr>  ) "bsr")
        ((= opcode <jmp>  ) "jmp")
        ((= opcode <jsr>  ) "jsr")
        ((= opcode <pea>  ) "pea")
        ((= opcode <addl> ) "addl")
        ((= opcode <addw> ) "addw")
        ((= opcode <cmpl> ) "cmpl")
        ((= opcode <subl> ) "subl")
        ((= opcode <moveq>) "moveq")
        ((= opcode <lea>  ) "lea")
        ((= opcode <addql>) "addql")
        ((= opcode <subql>) "subql")
        (else
         (compiler-internal-error "opcode->string, unknown 'opcode'" opcode))))

(define (make-code-block lbl-counter)
  (vector '()           ; 0 - frame size
          '()           ; 1 - temp cache
          (queue-empty) ; 2 - code queue
          (queue-empty) ; 3 - const queue
          0             ; 4 - length of the const queue
          (queue-empty) ; 5 - deferred code generation thunks
          lbl-counter)) ; 6 - label counter

(define (code-block-frame-size cb)        (vector-ref cb 0))
(define (code-block-frame-size-set! cb x) (vector-set! cb 0 x))
(define (code-block-temp-cache cb)        (vector-ref cb 1))
(define (code-block-temp-cache-set! cb x) (vector-set! cb 1 x))
(define (code-block-code-list cb)         (queue->list (vector-ref cb 2)))
(define (code-block-const-list cb)        (queue->list (vector-ref cb 3)))

(define (new-code-block-lbl! cb)
  ((vector-ref cb 6)))

(define (append-instr! cb instr)
  (let ((code-queue (vector-ref cb 2)))
    (if *asm-output?*
      (begin
        (for-each (lambda (x) (display x *asm-port*)) instr)
        (newline *asm-port*)))
    (queue-put! code-queue instr)))

(define (defer-append! cb thunk)
  (queue-put! (vector-ref cb 5) thunk))

(define (append-deferred! cb)
  (let ((thunk-list (queue->list (vector-ref cb 5))))
    (vector-set! cb 5 (queue-empty))
    (if (not (null? thunk-list))
      (let loop ((l thunk-list))
        (if (null? l)
          (append-deferred! cb)
          (begin
            (if *asm-output?* (newline *asm-port*))
            ((car l))
            (loop (cdr l))))))))

(define (append-instr-ea! cb opcode opnd1)
  (instruction-begin! cb)
  (let ((code-queue (vector-ref cb 2))
        (opnd1 (simulate-opnd! cb opnd1)))
    (if *asm-output?* (asm-write tab-char (opcode->string opcode) tab-char))
    (queue-put! code-queue (code-word (+ opcode (opnd->mode/reg opnd1))))
    (append-opnd! cb opnd1)
    (if *asm-output?* (asm-newline))))

(define (append-instr-ea-ea! cb opcode opnd1 opnd2)
  (instruction-begin! cb)
  (let* ((code-queue (vector-ref cb 2))
         (opnd1 (simulate-opnd! cb opnd1))
         (opnd2 (simulate-opnd! cb opnd2)))
    (if *asm-output?* (asm-write tab-char (opcode->string opcode) tab-char))
    (queue-put! code-queue
      (code-word (+ opcode
                    (opnd->mode/reg opnd1)
                    (* 512 (opnd->reg/mode opnd2)))))
    (append-opnd! cb opnd1)
    (if *asm-output?* (asm-write ","))
    (append-opnd! cb opnd2)
    (if *asm-output?* (asm-newline))))

(define (append-branch! cb opcode lbl)
  (let ((code-queue (vector-ref cb 2)))
    (if *asm-output?*
      (begin
        (asm-write tab-char (opcode->string opcode) tab-char "l" lbl)
        (asm-newline)))
    (queue-put! code-queue (code-branch-rel opcode lbl))))

(define (instruction-begin! cb)
  (set! *start-frame-size* (code-block-frame-size cb)))

(define (simulate-opnd! cb opnd)
  (cond ((identical-opnds? opnd pop-opnd)
         (code-block-frame-size-set! cb (- (code-block-frame-size cb) 1))
         opnd)
        ((identical-opnds? opnd push-opnd)
         (code-block-frame-size-set! cb (+ (code-block-frame-size cb) 1))
         opnd)
        ((and (not (= *start-frame-size* (code-block-frame-size cb)))
              (disp*? opnd)
              (identical-opnds? (disp*-areg opnd) sp-reg))
         (make-disp* sp-reg
                     (- (disp*-offset opnd)
                        (* 4 (- *start-frame-size*
                                (code-block-frame-size cb))))))
        (else
         opnd)))

(define (append-opnd! cb opnd)
  (if *asm-output?* (asm-write (opnd->string opnd)))
  (cond ((disp? opnd)
         (queue-put! (vector-ref cb 2)
           (code-word (disp-offset opnd))))
        ((rel? opnd)
         (queue-put! (vector-ref cb 2)
           (code-16bit-rel (rel-offset opnd) (rel-lbl opnd))))
        ((imm? opnd)
         (let ((code-queue (vector-ref cb 2)) (val (imm-val opnd)))
           (queue-put! code-queue
             (code-word (modulo (quotient val 65536) 65536)))
           (queue-put! code-queue
             (code-word (modulo val 65536)))))))

(define (append-const! cb val)
  (let ((const-queue (vector-ref cb 3)))
    (let loop ((l (queue->list const-queue)) (i 0))
      (cond ((null? l)
             (let ((num (vector-ref cb 4)))
               (vector-set! cb 4 (+ num 1))
               (queue-put! const-queue val)
               num))
            ((identical-objects? (car l) val)
             i)
            (else
             (loop (cdr l) (+ i 1)))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define const-lbl (+ label-limit 1))
(define end-lbl   (+ label-limit 2))
(define *temp-cache-on-exit* '())
(define *cb* '())
(define *entry-lbl* '())

(define (target-generate-procedure bb-set entry-lbl)
  (set! *entry-lbl* entry-lbl)
  (set! *cb* (make-code-block (basic-block-set-lbl-counter bb-set)))
  (set! *temp-cache-on-exit* '())

  (generate-basic-block-list! (basic-block-set->list bb-set entry-lbl))
  (append-deferred! *cb*)

  (display "------------------------------------------") (newline)
  (for-each (lambda (x) (for-each display x) (newline))
            (code-block-code-list *cb*))
  (display const-lbl) (display "$:") (newline)
  (for-each (lambda (x) (display "	.long ") (write x) (newline))
            (code-block-const-list *cb*))
  (display "	.align 4") (newline)
  (display end-lbl) (display "$:") (newline)
  (display "------------------------------------------") (newline)
)

(define (target-basic-block-begin! bb)
  (code-block-frame-size-set! *cb* (basic-block-frame-size bb))
  (code-block-temp-cache-set! *cb* empty-temp-cache))

(define (target-basic-block-end! bb)
  (set! *temp-cache-on-exit*
    (cons (cons bb (code-block-temp-cache *cb*)) *temp-cache-on-exit*)))

(define (target-generate-vm-instr! vm-instr)
  (let ((name (car vm-instr)) (parms (cdr vm-instr)))

    (if *vm-output?*
      (begin
        (if *asm-output?*
          (begin (newline *asm-port*) (display "| " *asm-port*))
          (if (not (eq? name 'LABEL)) (display "  " *asm-port*)))
        (write-vm-instr vm-instr *asm-port*)))

    (cond

      ((eq? name 'LABEL)
       (let ((lbl (car parms))
             (nb-parms (cadr parms))
             (rest? (caddr parms))
             (closed? (cadddr parms)))
         (if nb-parms
           (begin
             (append-instr! *cb* (list "	.align 4"))
             (if (= lbl *entry-lbl*)
               (append-instr! *cb* (list "	.word ("
                                         end-lbl
                                         "$-.)/4+procedure_type"))
               (append-instr! *cb* (list "	.word (.-"
                                         *entry-lbl*
                                         "$)/4+label_type")))
             (append-instr! *cb* (list lbl "$:"))
             (if rest?
               (begin
                 (append-instr! *cb*
                   (if closed?
                     (list "	jsr	a5@(rest_closed_label:w)")
                     (list "	jsr	a5@(rest_label:w)")))
                 (append-instr! *cb* (list "	.word " nb-parms)))
               (let ((trap (new-code-block-lbl! *cb*)))
                 (cond ((= nb-parms 1)
                        (append-instr! *cb* (list "	bpl	" trap "$")))
                       ((= nb-parms 2)
                        (append-instr! *cb* (list "	bne	" trap "$")))
                       (else
                        (append-instr! *cb*
                          (list (if (< nb-parms 8) "	subqw" "	subw")
                                "	#" (+ nb-parms 1) ",d0"))
                        (append-instr! *cb* (list "	bne	" trap "$"))))
                 (defer-append! *cb*
                   (lambda ()
                     (append-instr! *cb* (list trap "$:"))
                     (append-instr! *cb* (list "	jsr	a5@"))
                     (append-instr! *cb* (list "	.word (.-"
                                               lbl
                                               "$)/4+(arg_trap*4096)")))))))
           (append-instr! *cb* (list lbl "$:")))))

      ((eq? name 'CHECK_STACK)
       (let ((no-trap (new-code-block-lbl! *cb*)))
         (append-instr! *cb* (list "	cmpl	a6@,sp"))
         (append-instr! *cb* (list "	bcc	" no-trap "$"))
         (append-instr! *cb* (list "	jsr	a5@"))
         (append-instr! *cb* (list "	.word (.-"
                                   *entry-lbl*
                                   "$)/4+(stack_trap*4096)"))
         (append-instr! *cb* (list no-trap "$:"))))

      ((eq? name 'DEALLOC)
       (let ((n (car parms)))
         (let ((m (* n 4)))
           (cond ((<= m 8)    ; ==> addql #n*4,sp
                  (append-instr! *cb* (list "	addql	#" m ",sp")))
                 ((< m 32768) ; ==> lea sp@(n*4),sp
                  (append-instr! *cb* (list "	lea	sp@(" m "),sp")))
                 (else        ; ==> addl #n*4,sp
                  (append-instr! *cb* (list "	addl	#" m ",sp")))))))

      ((eq? name 'OPEN_CODE)
       (let ((prim (car parms))
             (opnds (cadr parms))
             (loc (caddr parms)))
         (append-instr! *cb* (list "..."))
         ))

      ((eq? name 'COPY)
       (let ((vm-opnd (car parms))
             (vm-loc (cadr parms)))
         (append-instr! *cb* (list "..."))
         ))

      ((eq? name 'MAKE_CLOSURES)
       (let ((sizes-locs (car parms)))
         (append-instr! *cb* (list "..."))
         ))

      ((or (eq? name 'IF) (eq? name 'IF_NOT))
       (let ((test (car parms))
             (vm-opnd1 (cadr parms))
             (vm-opnd2 (caddr parms))
             (lbl (cadddr parms)))
         (append-instr! *cb* (list "..."))
         ))

      ((eq? name 'JUMP)
       (let ((vm-opnd (car parms))
             (n (cadr parms))
             (nb-args (caddr parms))
             (check? (cadddr parms)))
         (append-instr! *cb* (list "..."))
         (append-deferred! *cb*)
         ))

      (else
       (compiler-internal-error
         "target-generate-vm-instr!, unknown 'name'"
         name)))))

; (encode-vm-opnd! cb vm-opnd source?) returns the MC68020 operand
; corresponding to the VM operand 'vm-opnd'.  'source?' indicates if the
; operand is used as a source or as a destination.  As a side effect,
; 'encode-vm-opnd' might append instructions to the code queue in 'cb' and
; change  the frame size in 'cb'.

(define (encode-vm-opnd! cb vm-opnd source?)
  (cond ((reg? vm-opnd)
         (let ((n (reg-num vm-opnd)))
           (if (= n 0) (make-areg 0) (make-dreg n))))
        ((stk? vm-opnd)
         (make-disp* sp-reg
                     (* 4 (- (code-block-frame-size cb) (stk-num vm-opnd)))))
        ((push/pop? vm-opnd)
         (if source? (make-pinc sp-reg) (make-pdec sp-reg)))
        ((glob? vm-opnd)
         (make-disp gv-reg (+ (* 6 (glob-num vm-opnd)) 256)))
        ((mem? vm-opnd)
         (make-disp* (bring-in-areg! cb (mem-base vm-opnd))
                     (mem-offset vm-opnd)))
        ((lbl? vm-opnd)
         )
        ((prim? vm-opnd)
         )
        ((obj? vm-opnd)
         (let ((val (obj-val vm-opnd)))
           (if (memory-allocated? val)
             (make-rel const-lbl (append-const! cb val))
             (make-imm (bit-representation val)))))
        (else
         (compiler-internal-error
           "encode-vm-opnd, unknown 'vm-opnd'"
           vm-opnd))))

; (bring-in-areg! cb vm-opnd) appends to 'cb' the code to load the operand
; 'vm-opnd' into an address register if it isn't already in one.  The address
; register containing 'vm-opnd' is returned.

(define (bring-in-areg! cb vm-opnd)
  (let ((areg (find-in-temp-cache! cb vm-opnd)))
    (or areg
        (let ((opnd (encode-vm-opnd! cb vm-opnd #t)))
          (if (areg? opnd)
            opnd
            (load-in-lru-temp! cb opnd vm-opnd))))))

(define (find-in-temp-cache! cb vm-opnd)
  (define (find vm-opnd l)
    (if (null? l)
      #f
      (let ((x (car l)))
        (if (identical-vm-opnds? vm-opnd (cdr x))
          l
          (let ((f (find vm-opnd (cdr l))))
            (and f (cons (car f) (cons x (cdr f)))))))))
  (let ((f (find vm-opnd (code-block-temp-cache cb))))
    (and f (begin (code-block-temp-cache-set! cb f) (caar f)))))

(define (load-in-lru-temp! cb opnd vm-opnd)
  (define (find l)
    (if (null? (cdr l))
      l
      (let ((f (find (cdr l))))
        (cons (car f) (cons (car l) (cdr f))))))
  (let* ((f (find (code-block-temp-cache cb)))
         (areg (caar f)))
    (code-block-temp-cache-set! cb
      (cons (cons areg (if (mem? vm-opnd) '() vm-opnd)) (cdr f)))
    (append-instr-ea-ea! cb <movl> opnd areg)
    areg))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (target-begin!) ; initialize package
  (set! *asm-port* (current-output-port))
  (set! *asm-output?* #t)
  (set! *vm-output?* #t))

(define (target-end!) ; finalize package
  '())

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Stuff local to the package:

(define empty-temp-cache
   (list (cons (make-areg 3) '()) 
         (cons (make-areg 2) '())
         (cons (make-areg 1) '())))

(define *start-frame-size* '())

(define tab-char (integer->char 9))

(define *asm-port* '())
(define *asm-output?* '())
(define *vm-output?* '())

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

;------------------------------------------------------------------------------


; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; definition of elements in code list:
;
; >= 0 : word
; < 0  : label
; ()   : align 4
; (lbl . disp) : pc relative offset with disp

(define (code-word x)              x)
(define (code-lbl lbl)             (- lbl))
(define (code-align4)              '())
(define (code-parent-rel kind)     (cons 0 kind))
(define (code-branch-rel kind lbl) (cons (- lbl) kind))
(define (code-16bit-rel disp lbl)  (cons lbl disp))

(define (code-word? x)       (and (integer? x) (>= x 0)))
(define (code-label? x)      (and (integer? x) (< x 0)))
(define (code-align4? x)     (null? x))
(define (code-parent-rel? x) (and (pair? x) (= (car x) 0)))
(define (code-branch-rel? x) (and (pair? x) (< (car x) 0)))
(define (code-16bit-rel? x)  (and (pair? x) (> (car x) 0)))

(define (assemble proc-type code-list const-list)

  (define header-offset 2)

  (define (build-fix-list len l)
    (if (null? l)
      '()
      (let ((x (car l)))
        (cond ((or (code-lbl? x) (code-align4? x))
               (cons (cons len l) (build-fix-list 0 (cdr l))))
              ((code-branch-rel? x)
               (cons (cons len l) (build-fix-list 2 (cdr l))))
              (else
               (build-fix-list (+ len 2) (cdr l)))))))

  (let ((fix-list (build-fix-list 0 code-list)))

    (define (label->loc lbl)
      (let loop ((loc header-offset) (l fix-list))
        (let ((x (cadar l)) (cur-loc (+ loc (caar l))))
          (cond ((eqv? x lbl)
                 cur-loc)
                ((code-align4? x)
                 (loop (* 4 (quotient (+ cur-loc 3) 4)) (cdr l)))
                (else
                 (loop cur-loc (cdr l)))))))

    (let again ()
      (let loop ((loc header-offset) (changed? #f) (l fix-list))
        (if (null? l)
          (if changed? (again))
          (let ((x (cadar l)) (cur-loc (+ loc (caar l))))
            (cond ((code-align4? x)
                   (loop (* 4 (quotient (+ cur-loc 3) 4)) changed? (cdr l)))
                  ((and (code-branch-rel? x)
                        (let ((dist (- (label->loc (car x)) (+ cur-loc 2))))
                          (or (< dist -128) (> dist 127) (= dist 0))))
                   (set-car! (cdar l) (cdr x)) ; convert short branch to long
                   (set-cdr! (cdar l) (cons (cons (abs (car x)) 0) (cddar l)))
                   (set-car! (car l) (+ (caar l) 2))
                   (set-cdr! (car l) (cddar l))
                   (loop (+ cur-loc 2) #t (cdr l)))
                  (else
                   (loop cur-loc changed? (cdr l))))))))

    (let loop ((loc header-offset) (l1 '()) (l2 code-list))
      (if (null? l2)
        (begin
          (for-each
            (lambda (l) (write-code tab ".word" tab (list-of l ",")))
            (partition (reverse l1) 10))
          (for-each
            (lambda (l) (write-code tab ".long" tab (list-of l ",")))
            (partition const-list 10)))
        (let ((x (car l2)))
          (cond ((code-word? x)
                 (loop (+ loc 2) (cons x l1) (cdr l2)))
                ((code-lbl? x)
                 (loop loc l1 (cdr l2)))
                ((code-align4? x)
                 (if (= (* 4 (quotient loc 4)) loc)
                   (loop loc l1 (cdr l2))
                   (loop (+ loc 2) (cons 0 l1) (cdr l2))))
                ((code-parent-rel? x)
                 (let ((dist (quotient loc 4)))
                   (if (> dist 4095)
                     (compiler-limitation-error
                      "procedure is too long [procedure header is too far]")
                     (loop (+ loc 2) (cons (+ (cdr x) dist) l1) (cdr l2)))))
                ((code-branch-rel? x)
                 (let ((dist (- (label->loc (car x)) (+ loc 2))))
                   (loop (+ loc 2)
                         (cons (+ (cdr x) (modulo dist 256)) l1)
                         (cdr l2))))
                ((code-16bit-rel? x)
                 (let ((dist (+ (- (label->loc (- (car x))) loc) (cdr x))))
                   (if (or (< dist -32768) (> dist 32767))
                     (compiler-limitation-error
                      "procedure is too long [label is too far]")
                     (loop (+ loc 2)
                           (cons (modulo dist 65536) l1)
                           (cdr l2)))))
                (else
                 (compiler-internal-error "assemble, unknown 'x'" x))))))))

(define (list-of l sep)
  (let loop ((l1 l) (l2 '()))
    (if (null? l1)
      (reverse l2)
      (loop (after-first l1)
            (cons (if (null? (after-first l1)) '() sep)
                  (cons (first l1)
                        l2))))))

(define (partition l n)
  (if (null? l)
    '()
    (let loop ((l l) (x '()) (y '()) (i (+ n 1)))
      (cond ((null? l)
             (reverse (cons (reverse y) x)))
            ((= i 1)
             (loop (after-first l)
                   (cons (reverse y) x)
                   (list (first l))
                   n))
            (else
             (loop (after-first l)
                   x
                   (cons (first l) y)
                   (- i 1)))))))

